# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import wx
from . import idmanager
from . import keys

class Window(object):
    """Frame class.

    Attributes:

    - wx
    - keys
    - menu
    """


    def __init__(self,
                 parent = None,
                 id = -1,
                 title = 'Title',
                 pos = wx.DefaultPosition,
                 size = wx.DefaultSize,
                 style = wx.DEFAULT_FRAME_STYLE | wx.CLIP_CHILDREN):
        """Create a Frame instance.
        """

        # Create the actual window (a wx Frame)
        self._wx = wx.Frame(parent, id, title, pos, size, style)

        self._keys = keys.Keys()
        self._keys.attach(self)

        # Set up the wx main menu bar
        self._menu_id_manager = idmanager.IDManager(10)
        menubar = wx.MenuBar()
        self._wx.SetMenuBar(menubar)
        self._menu = None

        self.CreateStatusBar()

        wx.EVT_LEFT_DOWN(self, self.onLeftDown)
        wx.EVT_LEFT_UP(self, self.onLeftUp)
        wx.EVT_MIDDLE_DOWN(self, self.onMiddleDown)
        wx.EVT_MIDDLE_UP(self, self.onMiddleUp)
        wx.EVT_RIGHT_DOWN(self, self.onRightDown)
        wx.EVT_RIGHT_UP(self, self.onRightUp)

    def popupMenu(self, menu, pos):
        """Open a menu as a popup menu at the given position.

        \param menu (\c Menu)  Menu tree
        \param pos (\c 2-sequence) Position of the popup menu
        """
        menu._attach(self._menu_id_manager, self, None)
        x,y = pos
        self.PopupMenuXY(menu.wx, x, y)
        menu._detach()

    def onLeftDown(self, event):
        event.Skip()

    def onLeftUp(self, event):
        event.Skip()

    def onMiddleDown(self, event):
        event.Skip()

    def onMiddleUp(self, event):
        event.Skip()

    def onRightDown(self, event):
        event.Skip()

    def onRightUp(self, event):
        event.Skip()


    ######################################################################
    ## protected:

    def __getattr__(self, name):
        """Forward attribute requests to the wxFrame object."""
        
        val = getattr(self._wx, name, None)
        if val==None:
            raise AttributeError("'Window' object has no attribute '%s'."%name)
        else:
            return val

        
    # "wx" property...
    
    def _getWx(self):
        """Return the corresponding wxFrame object.

        This method is used for retrieving the \a wx property.

        \return Frame (\c wxFrame).
        """
        return self._wx

    wx = property(_getWx, None, None, "wxFrame object")


    # "keys" property...
    
    def _getKeys(self):
        """Return the key manager.

        This method is used for retrieving the \a keys property.

        \return Key manager (\c Keys).
        """
        return self._keys

    keys = property(_getKeys, None, None, "Key manager")


    # "menu" property...
    
    def _getMenu(self):
        """Return the menu tree.

        This method is used for retrieving the \a menu property.

        \return Root menu node (\c MenuNode).
        """
        return self._menu

    def _setMenu(self, menu):
        """Set a menu.

        This method is used for setting the \a menu property.
        """

        if self._menu!=None:
            self._menu._detach()

        self._menu = menu
        self._menu._attach(self._menu_id_manager, self, self._wx.GetMenuBar())


    menu = property(_getMenu, _setMenu, None, "Menu tree")





# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: flockofbirds.py,v 1.2 2005/07/07 08:24:57 mbaas Exp $

## \file flockofbirds.py
## Contains the FlockOfBirds class.

import types
from . import component
from . import eventmanager, events
from .slots import *
from .cgtypes import *
from . import fob

# FlockOfBirds
class FlockOfBirds(component.Component):
    """Receives tracker values from a Flock of Birds motion tracker.

    For each bird the component creates the following four slots:

    - \c pos<n>_slot (\c Vec3Slot) - Position
    - \c angle<n>_slot (\c Vec3Slot) - Euler angles
    - \c matrix<n>_slot (\c Mat3Slot) - Rotation matrix
    - \c quat<n>_slot (\c QuatSlot) - Quaternion

    where <n> is the number of the bird (1-based). Not all of the slots
    are active at the same time. You can select which slot should carry
    the corresponding value via the bird mode (see constructor).

    """

    def __init__(self,
                 name = "FlockOfBirds",
                 com_port = 0,
                 baud_rate = 115200,
                 timeout = 2.0,
                 num_birds = 1,
                 bird_mode = "p",
                 hemisphere = "forward",
                 auto_insert = True):
        """Constructor.

        The bird mode is one of "p" (position), "a" (angles), "m" (matrix),
        "q" (quaternion) or a combination like "pa", "pm" and "pq".
        The argument \a bird_mode can either be a mode string that will
        be used for all birds, or it can be a list of mode strings for
        every bird.

        \param name (\c str) Component name
        \param com_port (\c int) COM port to use for communicating with the flock (0,1,...)
        \param baud_rate (\c int) Baud rate
        \param timeout (\c float) Time out value in seconds for RS232 operations
        \param num_birds (\c int) Number of birds to use (including the ERC)
        \param bird_mode (\c str or \c list) Tracker mode (selects what kind of data the birds will send)
        \param hemisphere (\c str) Hemisphere setting for the transmitter ("forward", "aft", "upper", "lower", "left", "right")
        \param auto_insert (\c bool) True if the component should be inserted into the scene automatically
        """
        
        component.Component.__init__(self, name, auto_insert)

        self.com_port = com_port
        self.baud_rate = baud_rate
        self.timeout = timeout
        self.num_birds = num_birds
        self.hemisphere = hemisphere

        if isinstance(bird_mode, str):
            self.bird_mode = self.num_birds*[bird_mode]
        else:
            self.bird_mode = bird_mode

        self.bird_mode = [x.upper() for x in self.bird_mode]

        if len(self.bird_mode)!=self.num_birds:
            raise ValueError("%d bird modes expected (got %d)"%(self.num_birds, len(self.bird_mode)))

        # Create slots
        self._slots = []
        for i in range(self.num_birds):
            p = Vec3Slot()
            a = Vec3Slot()
            m = Mat3Slot(mat3(1))
            q = QuatSlot(quat(1))
            self._slots.append((p,a,m,q))
            setattr(self, "pos%d_slot"%(i+1), p)
            setattr(self, "angle%d_slot"%(i+1), a)
            setattr(self, "matrix%d_slot"%(i+1), m)
            setattr(self, "quat%d_slot"%(i+1), q)
            self.addSlot("pos%d"%(i+1), p)
            self.addSlot("angle%d"%(i+1), a)
            self.addSlot("matrix%d"%(i+1), m)
            self.addSlot("quat%d"%(i+1), q)

        # Initialize the FOB...
        
        self._fob = fob.FOB()
        self._fob.open(self.com_port, self.baud_rate, self.timeout)
        
        self._fob.fbbAutoConfig(numbirds=self.num_birds)
        self._fob.fbbReset()
        hs = { "forward":fob.FORWARD,
               "aft":fob.AFT,
               "upper":fob.UPPER,
               "lower":fob.LOWER,
               "left":fob.LEFT,
               "right":fob.RIGHT }
        for i in range(1, self.num_birds):
            self._fob.hemisphere(hs[self.hemisphere.lower()], addr=i+1)
            self.setTrackerMode(self.bird_mode[i], i+1)
        self._fob.run()
        
        eventmanager.eventManager().connect(events.STEP_FRAME, self)
        

    ## protected:

    def onStepFrame(self):
        # Poll values
        for i in range(1, self.num_birds):
            values = self._fob.point(i+1)
            if values==None:
                print(("No values received from bird %d."%(i+1)))                
            else:
                # 1. Convert values into proper types
                #    (position, angles, matrix, quat)
                # 2. Set the values on the corresponding slots
#                print values
                ps,angs,ms,qs = self._slots[i]
                m = self.bird_mode[i]
                if m[0]=="P":
                    p = vec3(values[:3])
                    ps.setValue(p)
                c = m[-1]                    
                if c=="A":
                    a = vec3(values[-3:])
                    angs.setValue(a)
                elif c=="M":
                    m = mat3(values[-9:])
                    ms.setValue(m)
                elif c=="Q":
                    q = quat(values[-4:])
                    qs.setValue(q)
                    


    # setTrackerMode
    def setTrackerMode(self, mode, addr):
        """Issue the appropriate fob command to set the tracker mode.

        \param mode (\c str) Tracker mode ("P", "A", "M", "Q", "PA", "PM", "PQ")
        \param addr (\c int) Bird address
        """
        if mode=="P":
            self._fob.position(addr)
        elif mode=="A":
            self._fob.angles(addr)
        elif mode=="M":
            self._fob.matrix(addr)
        elif mode=="Q":
            self._fob.quaternion(addr)
        elif mode=="PA":
            self._fob.position_angles(addr)            
        elif mode=="PM":
            self._fob.position_matrix(addr)
        elif mode=="PQ":
            self._fob.position_quaternion(addr)
        else:
            raise ValueError("Invalid tracker mode: %s"%mode)
        



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: pyimport.py,v 1.4 2005/08/30 09:59:07 mbaas Exp $

import sys, os.path, copy
from . import pluginmanager
from . import globalscene
from . import eventmanager

# PyImporter
class PyImporter:

    _protocols = ["Import"]

    # extension
    def extension():
        return ["py"]
    extension = staticmethod(extension)

    # description
    def description(self):
        return "Python file"
    description = staticmethod(description)

    # importFile
    def importFile(self, filename):
        """Import a Python source file."""

        file_globals = {}
        # Use all globals from the cgkit package
        # Commented out the following line as is imported from cgkit.all now
#        file_globals = copy.copy(cgkit.__dict__)
        # Add some modules...
        exec("from cgkit.all import *", file_globals)
        exec("from cgkit.sl import *", file_globals)
        exec("from math import *", file_globals)
        # ...and some special global names...
        scene = globalscene.getScene()
        file_globals["scene"] = scene
        file_globals["timer"] = scene.timer()
        file_globals["worldroot"] = scene.worldRoot()
        file_globals["eventmanager"] = eventmanager.eventManager()
        
        paths = sys.path
        # Add the directory of the input file to the module search paths
        # so that local imports do work
        sys.path = [os.path.abspath(os.path.dirname(filename))] + sys.path
        # Import the file
        exec(compile(open(filename).read(), filename, 'exec'), file_globals)
        # Restore the old search paths
        sys.path = paths


######################################################################

# Register the OffImporter class as a plugin class
pluginmanager.register(PyImporter)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: events.py,v 1.3 2005/09/20 13:35:14 mbaas Exp $

## \file events.py
## Contains the standard event names and some predefined event classes.

import string
from . import keydefs

# STEP_FRAME is called whenever the timer is stepped forward one frame
# using timer.step().
# The event takes no arguments.
STEP_FRAME = "StepFrame"

# Reset the simulation/animation
RESET = "Reset"

# A key was pressed on the keyboard.
# The argument to the event is a KeyEvent object.
KEY_PRESS = "KeyPress"
# A key was released on the keyboard
# The argument to the event is a KeyEvent object.
KEY_RELEASE = "KeyRelease"

LEFT_DOWN = "LeftDown"
LEFT_UP = "LeftUp"
MIDDLE_DOWN = "MiddleDown"
MIDDLE_UP = "MiddleUp"
RIGHT_DOWN = "RightDown"
RIGHT_UP = "RightUp"
MOUSE_BUTTON_DOWN = "MouseButtonDown"
MOUSE_BUTTON_UP = "MouseButtonUp"
MOUSE_MOVE = "MouseMove"
MOUSE_WHEEL = "MouseWheel"

JOYSTICK_AXIS = "JoystickAxis"
JOYSTICK_BALL = "JoystickBall"
JOYSTICK_HAT = "JoystickHat"
JOYSTICK_BUTTON_DOWN = "JoystickButtonDown"
JOYSTICK_BUTTON_UP = "JoystickButtonUp"

SPACE_MOTION = "SpaceMotion"
SPACE_BUTTON_DOWN = "SpaceButtonDown"
SPACE_BUTTON_UP = "SpaceButtonUp"
SPACE_ZERO = "SpaceZero"

TABLET = "Tablet"

# KeyEvent
class KeyEvent:
    """Keyboard event (key press or release).

    This event is sent as argument to the KEY_PRESS and KEY_RELEASE events.
    The data is stored in the attributes key, keycode and mods.
    """
    
    def __init__(self, key, keycode, mods=0):
        """Constructor.

        \param key (\c unicode) Unicode key
        \param keycode Key (\c int) Key code (untranslated)
        \param mods (\c int) Modifier flags
        """
        # Unicode key
        self.key = key
        # Key code
        self.keycode = keycode
        # Modifier flags
        self.mods = mods

    def __str__(self):
        c = repr(self.key)
#        if self.key in string.printable:
#           c = self.key
#        else:
#            c = "."
        return "<KeyEvent key:%s (%d) mods:%d>"%(c,self.keycode,self.mods)

    # shiftKey
    def shiftKey(self):
        """Return True if the key is a Shift key."""
        return self.keycode==keydefs.KEY_SHIFT_LEFT or \
               self.keycode==keydefs.KEY_SHIFT_RIGHT

    # controlKey
    def controlKey(self):
        """Return True if the key is a Control key."""
        return self.keycode==keydefs.KEY_CONTROL_LEFT or \
               self.keycode==keydefs.KEY_CONTROL_RIGHT

    # altKey
    def altKey(self):
        """Return True if the key is an Alt key."""
        return self.keycode==keydefs.KEY_ALT_LEFT or \
               self.keycode==keydefs.KEY_ALT_RIGHT

# MouseButtonEvent
class MouseButtonEvent:
    """Mouse button event.
    """

    def __init__(self, button, x, y, x0, y0):
        """Constructor.

        \param button (\c int) Button number (1=Left / 2=Middle / 3=Right)
        \param x (\c int) Mouse position (pixel coordinate)
        \param y (\c int) Mouse position (pixel coordinate)
        \param x0 (\c float) Mouse position (normalized)
        \param y0 (\c float) Mouse position (normalized)
        """
        self.button = button
        self.x = x
        self.y = y
        self.x0 = x0
        self.y0 = y0

    def __str__(self):
        return "<MouseButtonEvent button:%d x:%d y:%d x0:%1.2f y0:%1.2f>"%(self.button, self.x, self.y, self.x0, self.y0)

# MouseWheelEvent
class MouseWheelEvent:
    """Mouse wheel event.
    """

    def __init__(self, delta, x, y, x0, y0):
        """Constructor.

        \param delta (\c int) Wheel delta
        \param x (\c int) Mouse position (pixel coordinate)
        \param y (\c int) Mouse position (pixel coordinate)
        \param x0 (\c float) Mouse position (normalized)
        \param y0 (\c float) Mouse position (normalized)
        """
        self.delta = delta
        self.x = x
        self.y = y
        self.x0 = x0
        self.y0 = y0

    def __str__(self):
        return "<MouseWheelEvent delta:%d x:%d y:%d x0:%1.2f y0:%1.2f>"%(self.delta, self.x, self.y, self.x0, self.y0)


# MouseMoveEvent
class MouseMoveEvent:
    """Mouse move event.
    """

    def __init__(self, x, y, dx, dy, x0, y0, dx0, dy0, buttons):
        """Constructor.

        \param x (\c int) Mouse position (pixel coordinate)
        \param y (\c int) Mouse position (pixel coordinate)
        \param dx (\c int) Mouse delta (pixel)
        \param dy (\c int) Mouse delta (pixel)
        \param x0 (\c float) Mouse position (normalized)
        \param y0 (\c float) Mouse position (normalized)
        \param dx0 (\c float) Mouse delta (normalized)
        \param dy0 (\c float) Mouse delta (normalized)
        \param buttons (\c int) Mouse buttons (each bit is a mouse button)
        """
        self.x = x
        self.y = y
        self.dx = dx
        self.dy = dy
        self.x0 = x0
        self.y0 = y0
        self.dx0 = dx0
        self.dy0 = dy0
        self.buttons = buttons

    def __str__(self):
        return "<MouseMoveEvent %d/%d (%d/%d) %1.2f/%1.2f (%1.3f/%1.3f) btns:%s>"%(self.x, self.y, self.dx, self.dy, self.x0, self.y0, self.dx0, self.dy0, hex(self.buttons))
#        return "<MouseMoveEvent x:%d y:%d dx:%d dy:%d x0:%1.2f y0:%1.2f dx:%1.2f dy:%1.2f buttons:%s>"%(self.x, self.y, self.dx, self.dy, self.x0, self.y0, self.dx0, self.dy0, hex(self.buttons))


# JoystickAxisEvent
class JoystickAxisEvent:
    """Joystick axis event.
    """

    def __init__(self, joystick, axis, value):
        """Constructor.

        \param joystick (\c int) Joystick ID (0,1,...)
        \param axis (\c int) Axis ID (0,1,...)
        \param value (\c float) The current value of the specified axis
        """
        self.joystick = joystick
        self.axis = axis
        self.value = value

    def __str__(self):
        return "<JoystickAxisEvent joystick:#%d axis:#%d value:%f>"%(self.joystick, self.axis, self.value)

# JoystickHatEvent
class JoystickHatEvent:
    """Joystick hat event.
    """

    def __init__(self, joystick, hat, x, y):
        """Constructor.

        \param joystick (\c int) Joystick ID (0,1,...)
        \param hat (\c int) Hat ID (0,1,...)
        \param x (\c int) The current x value of the specified hat
        \param y (\c int) The current x value of the specified hat
        """
        self.joystick = joystick
        self.hat = hat
        self.x = x
        self.y = y

    def __str__(self):
        return "<JoystickHatEvent joystick:#%d hat:#%d x:%d y:%d>"%(self.joystick, self.hat, self.x, self.y)

# JoystickBallEvent
class JoystickBallEvent:
    """Joystick ball event.
    """

    def __init__(self, joystick, ball, value):
        """Constructor.

        \param joystick (\c int) Joystick ID (0,1,...)
        \param ball (\c int) Ball ID (0,1,...)
        \param value (\c float) The current value of the specified ball
        """
        self.joystick = joystick
        self.ball = ball
        self.value = value

    def __str__(self):
        return "<JoystickBallEvent joystick:#%d ball:#%d value:%f>"%(self.joystick, self.ball, self.value)

# JoystickButtonEvent
class JoystickButtonEvent:
    """Joystick button event.

    This event is sent as argument to the JOYSTICK_BUTTON_DOWN and
    JOYSTICK_BUTTON_UP events.
    """

    def __init__(self, joystick, button):
        """Constructor.

        \param joystick (\c int) Joystick ID (0,1,...)
        \param button (\c int) Button ID (0,1,...)
        """
        self.joystick = joystick
        self.button = button

    def __str__(self):
        return "<JoystickButtonEvent joystick:#%d button:#%d>"%(self.joystick, self.button)


# SpaceMotion
class SpaceMotionEvent:
    """SpaceMotion event.

    This event is created when a SpaceMouse or SpaceBall is moved or rotated.
    """

    def __init__(self, translation, rotation, period):
        """Constructor.

        \param translation (\c vec3) Translation vector
        \param rotation (\c vec3) Rotation vector
        \param period (\c int) Time in milliseconds since the last event
        """
        self.translation = translation
        self.rotation = rotation
        self.period = period

    def __str__(self):
        return "<SpaceMotion t:%s r:%s period:%d>"%(self.translation, self.rotation, self.period)

# SpaceButton
class SpaceButtonEvent:
    """SpaceButton event.

    This event is generated when a SpaceMouse or SpaceBall button was
    pressed or released.
    """

    def __init__(self, button):
        """Constructor.
        
        \param button (\c int) Button number (1-29)
        """
        self.button = button

    def __str__(self):
        return "<SpaceButton button:%d>"%(self.button)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: lwobimport.py,v 1.1 2006/03/12 22:41:42 mbaas Exp $

import os.path, sys
from .cgtypes import *
from .worldobject import WorldObject
from .trimesh import TriMesh
from .trimeshgeom import TriMeshGeom
from .polyhedron import Polyhedron
from .polyhedrongeom import PolyhedronGeom
from . import pluginmanager
from . import cmds
from cgkit.all import UNIFORM, INT, GLMaterial
from . import lwob


class _LWOBReader(lwob.LWOBReader):
    """Read a Lightwave Object file (*.lwo).
    """
    
    def __init__(self, parent=None):
        lwob.LWOBReader.__init__(self)

        # Parent node for the Lightwave object
        self.parent = parent

        # The TriMeshGeom that receives the triangle mesh
        self.trimeshgeom = TriMeshGeom()
        # The PolyhedronGeom that receives the mesh (if it is no triangle mesh)
        self.polyhedrongeom = PolyhedronGeom()
        # The generated WorldObject
        self.worldobj = None
        # The number of surfaces in the file
        self.numsurfaces = 0
        # A mapping from surface name to material id
        # Key: Surface name / Value: Material id (0-based)
        self.surface_ids = {}

        # Message flags so that warning messages are only output once
        self.crv_msg = False
        self.patch_msg = False

    def handlePNTS(self, points):
        """Handle the points chunk.

        Stores the points in the TriMeshGeom.
        """
        verts = self.trimeshgeom.verts
        verts.resize(len(points))
        for i,p in enumerate(points):
            verts[i] = p
    
    def handleSRFS(self, names):
        """Handle the surface names chunk.
        """
        self.numsurfaces = len(names)
        for i,n in enumerate(names):
            self.surface_ids[n] = i
    
    def handlePOLS(self, polys):
        """Handle the polygons.

        This method creates the actual object. It is assumed that the
        points have been read before and are stored in the TriMeshGeom.
        It is also assumed that a SRFS chunk was present and numsurfaces
        is initialized.
        """
        # Assume the mesh is a triangle mesh and initialize the TriMeshGeom
        # first. If this fails, use a PolyhedronGeom instead...
        if self._initTriMesh(polys):
            geom = self.trimeshgeom
        else:
            # Copy the vertices into the polyhedron geom...
            numverts = self.trimeshgeom.verts.size()
            self.polyhedrongeom.verts.resize(numverts)
            self.trimeshgeom.verts.copyValues(0, numverts, self.polyhedrongeom.verts, 0)
            del self.trimeshgeom
            # Initialize the polys...
            self._initPolyhedron(polys)
            geom = self.polyhedrongeom

        w = WorldObject(name="lwob", parent=self.parent)
        w.setNumMaterials(self.numsurfaces)
        w.geom = geom
        self.worldobj = w

    def _initPolyhedron(self, polys):
        """Initialize the polys of the PolyhedronGeom.

        Sets the faces a poly mesh and adds a matid slot with the
        material indices.
        """
        geom = self.polyhedrongeom
        geom.setNumPolys(len(polys))
        geom.newVariable("matid", UNIFORM, INT)
        matid = geom.slot("matid")
        for i,(verts,surfid) in enumerate(polys):
            geom.setLoop(i, 0, verts)
            matid[i] = max(0, surfid-1)

    def _initTriMesh(self, polys):
        """Initialize the faces of the TriMeshGeom.

        Sets the faces of a triangle mesh and adds a matid slot with
        the material indices.
        If the mesh contains faces with more than 3 vertices the
        method aborts and returns False.
        """
        faces = self.trimeshgeom.faces
        faces.resize(len(polys))
        self.trimeshgeom.newVariable("matid", UNIFORM, INT)
        matid = self.trimeshgeom.slot("matid")
        for i,(verts,surfid) in enumerate(polys):
            if len(verts)!=3:
                self.trimeshgeom.deleteVariable("matid")
                return False
            faces[i] = verts
            matid[i] = max(0, surfid-1)
            
        return True
        

    def handleCRVS(self, curves):
        if not self.crv_msg:
            print("Curves are not yet supported.")
            self.crv_msg = True
    
    def handlePCHS(self, patches):
        if not self.patch_msg:
            print("Patches are not yet supported.")
            self.patch_msg = True
    
    def handleSURF(self, surface):
        """Handle a surface chunk.

        Currently this just creates a GLMaterial with the base color of
        the surface. Everything else is ignored so far.
        """
        if surface.name not in self.surface_ids:
            raise lwob.LWOBError('Invalid surface name "%s" (name not available in SRFS chunk)'%surface.name)
        
        id = self.surface_ids[surface.name]
        
        col = surface.color
        if col==None:
            col = (255,255,255)
            
        mat = GLMaterial(diffuse=vec3(col)/255)
        self.worldobj.setMaterial(mat, id)


# LWOBImporter
class LWOBImporter:

    _protocols = ["Import"]

    # extension
    def extension():
        """Return the file extensions for this format."""
        return ["lwo"]
    extension = staticmethod(extension)

    # description
    def description(self):
        """Return a short description for the file dialog."""
        return "Lightwave object file"
    description = staticmethod(description)

    # importFile
    def importFile(self, filename, parent=None):
        """Import an LWOB file."""

        f = file(filename, "rb")
        reader = _LWOBReader(parent=parent)
        reader.read(f)
        f.close()

######################################################################

# Register the Importer class as a plugin class
pluginmanager.register(LWOBImporter)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

## \file menu.py
## Contains the menu tree classes.

import wx
import types


class MenuNodeNotFound(Exception):
    """Exception class."""
    pass

class DuplicateNames(Exception):
    """Exception class."""
    pass

class InvalidMenuRoot(Exception):
    """Exception class."""
    pass

class NoWxMenuObjectFound(Exception):
    """Exception class."""
    pass

class MenuAlreadyInUse(Exception):
    """Exception class."""
    pass

class NoMenuRoot(Exception):
    """Exception class."""
    pass

NODE_NORMAL_ITEM  = 0x01
NODE_CHECK_ITEM   = 0x02
NODE_COMMAND_ITEM = NODE_NORMAL_ITEM | NODE_CHECK_ITEM
NODE_SEPARATOR    = 0x04
NODE_MENU         = 0x08
NODE_MARKER       = 0x10
NODE_ALL_NODES    = 0xff


# MenuNode
class MenuNode(object):
    """Base class for a menu node.

    This is the base class for all menu nodes (inner node or leaf).
    A %MenuNode class can be used as a sequence where the items are
    the children nodes. Leaf nodes are always empty.
    
    Each node has the following attributes:

    - \b name (\c str): Node name
    - \b absname (\c str): Absolute node name
    - \b parent (\c %MenuNode): Parent node
    - \b root (\c %MenuNode): Root node
    - \b text (\c str): Text that's visible to the user
    - \b enabled (\c bool): If this flag is false the menu or menu item is grayed out.
    - \b visible (\c bool): If this flag is false the menu or menu item isn't displayed at all.
    - \b wx (\c wxMenu, \c wxMenuBar or \c wxMenuItem): Corresponding wxPython object (or None if the menu isn't attached).
    
    - flags

    A menu node can be either in an \em attached or \em detached state
    which tells if a node is attached to a wxPython menu or not. An
    attached menu is "active" and usually has a corresponding wxPython
    menu object associated with it. When you create a menu node it's
    not attached to a menu. A node automatically becomes attached when
    it's added to another node that is already attached.

    Any class that is derived from this base class has to implement the
    following methods:

    - setEnabled()
    - setVisibility()
    - _wx_menu_count()
    - _create_wx()
    
    """
    
    def __init__(self, name=None, parent=None, text=None,
                 enabled=True, visible=True):
        """Constructor.

        If no name is given the constructor tries to create a name
        from the \a text argument (using _create_auto_name()).

        \param name (\c str) Node name
        \param parent (\c %MenuNode) Parent node
        \param text (\c str) %Menu title or item text
        \param enabled (\c bool) Initial enabled state
        \param visible (\c bool) Initial visibility state
        """
        # Node name
        self._name = name
        # Parent node
        self._parent = parent
        # Text (menu title or item text)
        self._text = text
        # NODE_xxx flags
        self._flags = 0
        # Enabled state
        self._enabled = enabled
        # Visbility state
        self._visible = visible
        # The corresponding wxPython object (or None)
        self._wx = None
        # The wxPython ID of the object
        self._id = None
        # Activation flag
        self._attached = False

        if self._name==None:
            self._create_auto_name(self._text)


    # isAttached
    def isAttached(self):
        return self._attached

    def setEnabled(self, flag=True):
        """Set the enabled flag of this node.

        If a node is disabled all of its children are considered disabled
        as well.

        This method should be overridden by derived classes. Those
        implementations should call the inherited method (where the
        internal flag _enabled is set). If the node is attached to a
        menu this method should update their corresponding menu items
        to reflect the state change.

        This method is called automatically when the \em enabled property
        receives a new value.

        \param flag (\c bool) New enabled state.
        """
        self._enabled = flag

    def setVisibility(self, flag=True):
        """Set the visibility flag of this node.

        If a node is invisible all of its children are considered invisible
        as well.

        This method should be overridden by derived classes. Those
        implementations should call the inherited method (where the
        internal flag _visible is set). If the node is attached to a
        menu this method should remove their corresponding menu items
        to reflect the state change.

        This method is called automatically when the \em visible property
        receives a new value.

        \param flag (\c bool) New visibility state.
        """
        
        self._visible = flag

        if flag:
            if not self.isAttached():
                self._create_wx()
        else:
            if self.isAttached():
                self._destroy_wx()


    # remove
    def remove(self, node=""):
        """Remove a menu node.

        Removes the node with the given name. The name may contain an
        entire (relative) path to a menu node or may be left blank if
        self should be removed (which is the default behavior). The node
        is removed from the visible wx menu and from the menu tree and
        is returned to the caller which may ignore it or use it to add
        it later or to another menu.

        \param node (\c str) Name of the node which should be removed (default: "").
        \return The removed menu node object (\c MenuNode).
        """

        # Find the node n which is to be removed
        n = self.findNode(node)

        # Remove the corresponding wx objects from the visible menu
        if n.isAttached():
            n._destroy_wx()

        parent = n.parent
        if parent!=None:
            del parent[n]

        return n

    def findCommandNodes(self, func):
        """Return all nodes that are bound to func.

        \return A list of all nodes that are bound to func.
        """

        res = []
        
        # Check if this node is bound to func
        f = getattr(self, "_command", None)
        if f!=None:
            if f==func:
                res.append(self)

        # Now check the children
        for n in self:
            res+=n.findCommandNodes(func)

        return res                
        

    # findNode
    def findNode(self, path, exc=True):
        """Search for a node with a specific relative name.

        The name can include an entire path. Each node name is separated
        by a dot. If path is empty then self is returned.

        \param path (\c str) Name of the item
        \param exc (\c bool) If True then an exception is generated when the
                             node is not found.
        \return Node or None/Exception.
        """

        if path=="":
            return self

        f = path.split(".")
        name = f[0]
        for n in self:
            if name==n.name:
                if len(f)==1:
                    return n
                else:
                    return n.findNode(".".join(f[1:]))

        if exc:
            errpath = name
            s = self.absname
            if s!="":
                errpath = ".".join([s,errpath])
            raise MenuNodeNotFound('Menu node "%s" not found.'%(errpath))
        else:
            return None
            

    ######################################################################
    ## protected:

    def __len__(self):
        return 0

    def __getitem__(self, key):
        if isinstance(key, int):
            raise IndexError("sequence index out of range")
        else:
            raise TypeError("sequence indices must be integers")

    # _create_auto_name
    def _create_auto_name(self, text):
        """Create a name from the given text.

        The argument \a text is the text that's visible to the user
        in the menu. The method tries to generate a node name out
        of the given text. If the text only consists of dashes or is None
        then no name is generated (None), otherwise blanks and tabs are
        collapsed and replaced by underscores, dots and ampersands are
        deleted and the text is made lower case to create the name.

        The result of the automatic name generation is stored in the
        name property (_name).
        
        \param text (\c str) Menu text.
        """
        if text==None:
            self._name = None
            return
        text = text.strip()
        # Collapse dash sequences
        while text.find("--")!=-1:
            text = text.replace("--", "-")
        # Was the text that of a separator? then no name is generated
        if text=="-":
            self._name = None
            return
        # Replace tabs with spaces
        text = text.replace("\t", " ")
        # Collapse spaces
        while text.find("  ")!=-1:
            text = text.replace("  ", " ")
        # Replace/delete special characters
        text = text.replace("&","")
        text = text.replace(".","")
        text = text.replace(" ","_")
        self._name = text.lower()

    def _wx_menu_count(self):
        """Return the number of menu "slots" that are occupied by this node.

        The return value is the number of attached menu items (or
        menus) that are beneath this node. Invisble nodes must not be
        counted. This value is used to determine the position of a
        node in the wx menu.

        By default this method returns 0. Usually, a derived class should
        override this method.

        \return Number of visible items occupied by this node.
        \see _wx_menu_pos()
        """
        return 0

    def _wx_menu_pos(self):
        """Return the position of the node inside the parent wx menu.

        The return value is the starting index where self is located
        (or would be located) in the corresponding wx menu. This value
        is used to insert the node into a menu or menubar.

        The node has to be added to the menu tree but doesn't have to
        be attached.
        
        \pre self is in the children list of parent.
        \pre The parent nodes must already be attached.
        \return Position (\c int) or None if the node doesn't have a parent.
        \see _wx_menu_count()
        """

        parent = self.parent
        if parent==None:
            return None

        # Determine the index of self within parent
        idx = 0
        for n in parent:
            if n==self:
                break
            idx += n._wx_menu_count()

        if parent._wx==None:
            idx += parent._wx_menu_pos()

        return idx

        

    def _attach(self, IDmanager, window, wxmenubar):
        """Attach the menu to a window.
        
        \param IDmanager ID generator
        \param window (\c Window) wxPython window
        \param wxmenubar (\c wxMenuBar) wxPython menu bar or None
        """
        root = self.root
        if not isinstance(root, Menu):
            raise InvalidMenuRoot('The root of a menu has to be a Menu object.')
        # The node must not be attached to another menu
        if self.isAttached():
            raise MenuAlreadyInUse("The menu node is already in use.")

        root._IDmanager = IDmanager
        root._window = window
        if wxmenubar!=None:
            root._wx_menubar = wxmenubar
        root._menu_id_lut = {}
        root._create_wx()

    def _detach(self):
        """Detach the menu from a window."""
        
        if self.root!=self:
            raise NoMenuRoot("Only root nodes can be detached.")

        self._destroy_wx()
        del self._IDmanager
        del self._window
        if hasattr(self, "_wx_menubar"):
            del self._wx_menubar
        del self._menu_id_lut
        

    def _create_wx(self):
        """Create the corresponding wxPython object.

        This method has to create the appropriate wxPython object which
        is represented by the node and insert it into the parent menu,
        submenu or menubar at the correct position as returned by
        _wx_menu_pos(). 
        
        The wxPython object has to be stored in the attribute self._wx
        and its ID in self._id.
        The method may only be called if the parent wxPython object
        of the node exists or if the node represents the entire menu bar.

        A derived class must call this inherited method!
        """
        self._attached = True

    def _destroy_wx(self):
        """Destroy the associated wxPython object.

        This method has to remove its corresponding wxPython object
        from the visible wx menu. The method may only be called in
        attached state!

        A derived class must call this inherited method!
        """
        self._attached = False
        self._wx = None
        self._id = -1        

    def _wx_parent(self):
        """Return the wxPython parent object.

        The object returned is either a wx.Menu, a wx.MenuBar or None.
        If a node has no direct wx parent, the wx parent of the
        parent node is returned.

        This method must not be called when the node is part of an
        invisible tree.

        \return wxPython object or None.
        """
        p = self.parent

        while p!=None:
            if p._wx!=None:
                if isinstance(p._wx, wx.MenuItem):
                    return p._wx.GetSubMenu()
                else:
                    return p._wx
            p = p.parent

        return None


    def _is_menubar(self):
        return False
    
    def _is_embedded_menu(self):
        return False


    # "name" property...
    
    def _getName(self):
        """Return the node name.

        This method is used for retrieving the \a name property.

        \return Node name (\c str).
        """
        return self._name

    name = property(_getName, None, None, "Node name")

    # "absname" property...

    def _getAbsName(self):
        """Return the absolute name of the node.

        Returns the absolute name of the node. This name
        uniquely identifies the node within its tree and is composed
        of all the names from the root to the node separated by a dot.
        The root node is always unnamed.

        This method is used for retrieving the \a absname property.

        \return Absolute name of the node (\c str).
        """
        if self._parent==None:
            return ""
        else:
            pre = self._parent.getFullName()
            if self._name==None:
                name = "<unnamed>"
            else:
                name = self._name
            if pre=="":
                return name
            else:
                return pre+"."+name

    absname = property(_getAbsName, None, None, "Absolute node name")

    # "parent" property...
    
    def _getParent(self):
        """Return the parent node.

        This method is used for retrieving the \a parent property.

        \return Parent node (\c MenuNode).
        """
        return self._parent

    parent = property(_getParent, None, None, "Parent node")

    # "root" property...

    def _getRoot(self):
        """Return the root node.

        This method is used for retrieving the \a root property.

        \return Root node (\c MenuNode).
        """
        p = self
        while 1:
            if p._parent==None:
                return p
            p = p._parent

    root = property(_getRoot, None, None, "Root node")

    # "text" property...
    
    def _getText(self):
        """Return the menu text which is displayed to the user.

        This method is used for retrieving the \a text property.

        \return Text (\c str).
        """
        return self._text

    text = property(_getText, None, None, "Menu text")

    # "enabled" property...
    
    def _getEnabled(self):
        """Return the enabled state.

        This method is used for retrieving the \a enabled property.

        \return True if the node is enabled.
        """
        return self._enabled

    def _setEnabled(self, flag=True):
        """Set the enabled state.

        This method is used for setting the \a enabled property.

        \param flag (\c bool) New enabled state.
        """
        self.setEnabled(flag)

    enabled = property(_getEnabled, _setEnabled, None, "Enabled state")

    # "visible" property...
    
    def _getVisible(self):
        """Return the visible state.

        This method is used for retrieving the \a visible property.

        \return True if the node is visible.
        """
        return self._visible

    def _setVisible(self, flag=True):
        """Set the visible state.

        This method is used for setting the \a visible property.

        \param flag (\c bool) New visibility state.
        """
        self.setVisibility(flag)

    visible = property(_getVisible, _setVisible, None, "Visibility state")

    # "wx" property...
    
    def _getWx(self):
        """Return the corresponding wxPython object.

        This method is used for retrieving the \a wx property.

        \return wxPython object.
        """
        return self._wx

    wx = property(_getWx, None, None, "Corresponding wxPython object")



######################################################################
######################################################################
        
# MenuMarker
class MenuMarker(MenuNode):
    """This class represents a position in a menu.
    """
    def __init__(self, name=None, parent=None):
        MenuNode.__init__(self, name=name, parent=parent)
        self._flags = NODE_MARKER

    def __repr__(self):
        return "<MenuMarker name=%s>"%(self.name)

######################################################################
######################################################################

# MenuItem
class MenuItem(MenuNode):
    """An individual menu item.

    This class represents one individual menu item which can be a
    normal item, a checkable item or a separator. A menu item object
    has the following attributes in addition to the ones inherited by
    MenuNode:

    - \b command (\c callable): The command to execute (may be None)
    - \b checked (\c bool): The state of a checkable menu item
    """
    
    def __init__(self, itemdef, name=None, enabled=True, visible=True):
        """Constructor.

        The definition \a itemdef of the item is generally composed of
        two items, the text that appears in the menu and the
        corresponding command which is to be executed whenever the
        user selects the menu item. The command must be a callable
        that takes no arguments.
        
        The text and command is given as a 2-tuple (text, command).
        It's also possible to just provide a string which is equivalent
        to (text, None).

        If the text starts with "[x]" then a checkable menu item is
        generated.

        If the text consists of 3 or more dashes ("---"), then a
        separator is generated.

        \param itemdef (\c 2-tuple or \c str) Item definition
        \param name (\c str) The internal name of the item
        \param enabled (\c bool) Initial enabled state
        \param visible (\c bool) Initial visibility state
        """

        flags = NODE_NORMAL_ITEM

        self._checkitem = False
        self._separator = False
        self._checked   = False

        # Split the item definition into the text and command component
        if isinstance(itemdef, (str,)):
            itemstr = itemdef
            itemcmd = None
        else:
            itemstr, itemcmd = itemdef

        # Is the menu item a checkable item?
        if itemstr[:3]=="[x]" or itemstr[:3]=="[X]":
            flags = NODE_CHECK_ITEM
            self._checkitem = True
            if itemstr[1]=="X":
                self._checked=True
            itemstr = itemstr[3:]

        # Is the item a separator?
        if len(itemstr)>2 and itemstr.count("-")==len(itemstr):
            flags = NODE_SEPARATOR
            self._separator = True
            itemcmd = None

        MenuNode.__init__(self, name=name, text=itemstr, enabled=enabled, visible=visible)
        self._flags = flags

        self._command   = itemcmd

    def __repr__(self):
        s = 'MenuItem name=%s text="%s"'%(self.name, self.text)
        if self._checkitem:
            s+=" checkable"
        return "<%s>"%s

    def isSeparator(self):
        return self._separator

    def isCheckItem(self):
        return self._checkitem


    def setEnabled(self, flag=True):
        MenuNode.setEnabled(self, flag)
        if self._wx!=None:
            self._wx.Enable(flag)

    def update(self):
        if self._wx!=None:
            root = self.root
            text = self.text
            keys = root._window.keys.findCommandKeys(self._command)
            if len(keys)>0:
                text+="\t"+keys[0]
            self._wx.SetText(text)


    ######################################################################

    ## protected:

    def _wx_menu_count(self):
        if self._wx==None:
            return 0
        else:
            return 1

    def _create_wx(self):

        MenuNode._create_wx(self)

        root = self.root

        # Determine the type of MenuItem to create...
        if self.isCheckItem():
            kind = wx.ITEM_CHECK
            self._id = root._IDmanager.allocID()
        elif self.isSeparator():
            kind = wx.ITEM_SEPARATOR
            self._id = wx.ID_SEPARATOR
        else:
            kind = wx.ITEM_NORMAL
            self._id = root._IDmanager.allocID()

        # Get the parent which must be a wxMenu object
        parent = self._wx_parent()
        # Create the MenuItem...
        text = self.text
        keys = root._window.keys.findCommandKeys(self._command)
        if len(keys)>0:
            text+="\t"+keys[0]
        self._wx = wx.MenuItem(parent, self._id, text, self._getHelpString(), kind)
        # ...and insert it into the parent menu
        pos = self._wx_menu_pos()
#        print 'Inserting item "%s" at position %d'%(self._text, pos)
        parent.InsertItem(pos, self._wx)
        # Set initial attributes
        self._wx.Enable(self.enabled)
        if self.isCheckItem():
            self._wx.Check(self.checked)

        # Connect the new wxMenuItem to the menu event handler
        if not self.isSeparator():
            root._connect(self)


    def _destroy_wx(self):

        root = self.root
        
        # Disconnect the menu item from the event handler
        if not self.isSeparator():
            root._disconnect(self)

        # Remove the wxMenuItem from the parent menu
        parent = self._wx_parent()
        parent.Remove(self._id)
        # Make the ID available again
        if self._id!=-1:
            root._IDmanager.freeID(self._id)

        MenuNode._destroy_wx(self)

    def _getHelpString(self):
        doc = getattr(self._command, "__doc__", None)
        if doc==None:
            return ""
        else:
            return doc.split("\n")[0]

    # "command" property...

    def _getCommand(self):
        """Return the associated command (callable object).

        This method is used for retrieving the \a command property.

        \return Command (\c callable).
        """
        return self._command

    def _setCommand(self, command):
        """Set a new command.

        This method is used for setting the \a command property.

        \param command (\c callable) The new command or None
        """
        self._command = command

    command = property(_getCommand, _setCommand, None, "Menu item command")

    # "checked" property...

    def _getChecked(self):
        """Get the checked flag.

        This method is used for retrieving the \a checked property.

        \return Checked flag (\c bool).
        """
        return self._checked
    
    def _setChecked(self, flag=True):
        """Get the checked flag.

        This method is used for setting the \a checked property.

        \param flag (\c bool) New state.
        """
        self._checked = flag
        if self._wx!=None:
            self._wx.Check(flag)

    checked = property(_getChecked, _setChecked, None, "Checked flag")


######################################################################
######################################################################
       
# Menu
class Menu(MenuNode):
    """A menu node containing other menus, markers or menu items.

    This class is used as a container for MenuNode classes and it
    represents a node in the entire menu tree.
    """
    
    def __init__(self, title=None, name=None, enabled=True, visible=True,
                 items=[]):
        """Constructor.

        \param title (\c str) Menu title which will be display in the GUI
        \param name (\c str) Node name
        \param enabled (\c bool) Initial enabled state
        \param visible (\c bool) Initial visibility state
        \param items (\c sequence) A sequence of children node descriptions
        """

        MenuNode.__init__(self, name=name, text=title, enabled=enabled, visible=visible)
        self._flags = NODE_MENU

        # Children nodes
        self._children = [MenuMarker("START", self), MenuMarker("END", self)]

        for item in items:
            if isinstance(item, MenuNode):
                mi = item
            else:
                mi = MenuItem(item)
            self.append(mi)


    def insertAfter(self, newnode, treenodename):
        """Insert a node after another existing node.

        \param newnode  The node that should be inserted
        \param treenodename The node name after which newnode is to be inserted
        """

        # The new node must not be attached to another menu
        if newnode.isAttached() or newnode.parent!=None:
            raise MenuAlreadyInUse("The menu node is already in use.")

        treenode = self.findNode(treenodename)
        # If treenode is not a direct child then delegate the insert
        # operation to the parent of treenode
        if (treenode.parent!=self):
            treenode.parent.insertAfter(newnode, treenodename.split(".")[-1])
            return

        # Check if the name of the new node collides with an existing node
        newnodename = newnode.name
        if newnodename!=None and self.findNode(newnodename, False)!=None:
            raise DuplicateNames('Menu node "%s" already exists.'%newnodename)

        # Add the new node into the children list
        idx = self._children.index(treenode)
        self._children.insert(idx+1, newnode)
        newnode._parent = self

        if self.isAttached():
            newnode._create_wx()


    def insertBefore(self, newnode, treenodename):
        """Insert a node before another existing node.

        \param newnode  The node that should be inserted
        \param treenodename The node name before which newnode is to be inserted
        """
        
        # The new node must not be attached to another menu
        if newnode.isAttached() or newnode.parent!=None:
            raise MenuAlreadyInUse("The menu node is already in use.")

        treenode = self.findNode(treenodename)
        # If treenode is not a direct child then delegate the insert
        # operation to the parent of treenode
        if (treenode.parent!=self):
            treenode.parent.insertBefore(newnode, treenodename.split(".")[-1])
            return

        # Check if the name of the new node collides with an existing node
        newnodename = newnode.name
        if newnodename!=None and self.findNode(newnodename, False)!=None:
            raise DuplicateNames('Menu node "%s" already exists.'%newnodename)

        # Add the new node into the children list
        idx = self._children.index(treenode)
        self._children.insert(idx, newnode)
        newnode._parent = self

        if self.isAttached():
            newnode._create_wx()

    # append
    def append(self, node):
        """Append a node at the end of the children list.

        The node is added right before the END marker.

        \param node (\c MenuNode) The node the be appended to the menu.
        """
        self.insertBefore(node, "END")

       

    def setEnabled(self, flag=True, node=""):
        if node!="":
            # Find the node n whose enabled state should be changed
            n = self.findNode(node)
            n.setEnabled(flag)
            return

        MenuNode.setEnabled(self, flag)
        if not self.isAttached():
            return

        # 1. Node = MenuBar
        if self._is_menubar():
            for n in self:
                n.setEnabled(flag)
        # 2. Node = Toplevel menu
        elif self.parent==None or self.parent._is_menubar():
            parent = self._wx_parent()
            # parent may be None if the menu is used for a popup menu
            if parent!=None:
                parent.EnableTop(self._wx_menu_pos(), flag)
        # 3. Node = Sub menu
        elif self.text!=None:
            parent = self._wx_parent()
            parent.Enable(self._wx.GetId(), flag)
        # 3. Node = Embedded menu
        else:
            for n in self:
                n.setEnabled(flag)              


    ######################################################################
    ## protected:

    def _wx_menu_count(self):
        if self._is_embedded_menu():
            res = 0
            for n in self:
                res += n._wx_menu_count()
            return res
        
        if self._wx==None:
            return 0
        else:
            return 1

    def __repr__(self):
        return '<Menu name=%s title="%s">'%(self.name, self.text)

    def __len__(self):
        return len(self._children)

    def __getitem__(self, key):
        return self._children[key]

    def __delitem__(self, key):
        # Remove the node from the tree
        self._children.remove(key)
        key._parent = None

    def __getattr__(self, name):
        items = [x for x in self._children if name==x.name]
        if items==[]:
            raise AttributeError('Menu node "%s" not found.'%(name))
        return items[0]

    def __setattr__(self, name, value):
        if name=="":
            return
        if name[0]=="_":
            MenuNode.__setattr__(self, name, value)
        elif isinstance(value, Menu):
            self.append(value)
        elif isinstance(value, MenuItem):
            self.append(value)
        elif (isinstance(value, (str,)) or
              (isinstance(value, tuple) and len(value)==2 and
              isinstance(value[0], (str,)) and callable(value[1]))):
            mi = MenuItem(value, name=name)
            self.append(mi)
        else:
            MenuNode.__setattr__(self, name, value)
            

##    def _getIndex(self, name):
##        """Return the index of the children with the given name.

##        \return Index (0-based).
##        """
        
##        items = filter(lambda x: name==x.name, self._children)
##        if items==[]:
##            raise MenuNodeNotFound, 'Menu node "%s" not found.'%(name)
##        return self._children.index(items[0])

    def _create_wx(self):

        MenuNode._create_wx(self)

        # 1. Node = MenuBar
        if self._is_menubar():
#            print "Menubar"
            self._create_wx_menubar()    
        # 2. Node = Toplevel menu
        elif self.parent==None or self.parent._is_menubar():
#            print "Toplevel"
            self._create_wx_toplevel_menu()
        # 3. Node = Sub menu
        elif self.text!=None:
#            print "Submenu"
            self._create_wx_submenu()
        # 3. Node = Embedded menu
        else:
#            print "Embedded"
            self._create_wx_embedded_menu()

        self.setEnabled(self.enabled)


    def _create_wx_menubar(self):
        """Implements _create_wx() for the root node.

        This method just calls _create_wx() on all its children.
        (The wxMenuBar object must already be set!)

        \see _create_wx()
        """
        # _wx_menubar must already be initialized with a wx.MenuBar object

        self._wx = self._wx_menubar
        
        # Create wx objects for all menus...
        for node in self:
            node._create_wx()


    def _create_wx_toplevel_menu(self):
        """Implements _create_wx() for toplevel menus.

        This method creates a wxMenu object and fills it with its
        children.

        \see _create_wx()
        """

        self._wx = wx.Menu()

        # Create wx objects for all items...
        for node in self:
            node._create_wx()

        # Parent is a wx.MenuBar object or None (popup menus)
        parent = self._wx_parent()
        if parent!=None:
            pos = self._wx_menu_pos()
            parent.Insert(pos, self._wx, self.text)


    def _create_wx_submenu(self):
        """Implements _create_wx() for submenus.

        This method creates a wxMenu object and fills it with its
        children. Then the wxMenu object is wrapped in a wxMenuItem
        object which serves as the actual menu item.

        \see _create_wx()
        """

        # Store the menu in _wx temporarily, so that the children have a
        # valid parent
        self._wx = wx.Menu()

        # Create wx objects for all items...
        for node in self:
            node._create_wx()

        # Parent is a wx.Menu object
        parent   = self._wx_parent()
        self._id = self.root._IDmanager.allocID()
        self._wx = wx.MenuItem(parent, self._id, self.text, "",
                               wx.ITEM_NORMAL, self._wx)
        pos = self._wx_menu_pos()
        parent.InsertItem(pos, self._wx)
       

    def _create_wx_embedded_menu(self):
        """Implements _create_wx() for embedded menus.

        This method does not create a wx menu object but only inserts
        its children into the parent menu.        

        \see _create_wx()
        """

        self._wx = None
        self._id = None
        wxmenu = self._wx_parent()
        # Create wx objects for all items...
        for node in self:
            node._create_wx()


    def _destroy_wx(self):

        # Destroy the wx objects for all items...
        for node in self:
            node._destroy_wx()

        # 1. Node = MenuBar
        if self._is_menubar():
            pass
        # 2. Node = Toplevel menu
        elif self.parent==None or self.parent._is_menubar():
            # Remove the menu from the menu bar (if it's not a popup menu)
            parent = self._wx_parent()
            if parent!=None:
                pos = self._wx_menu_pos()
                parent.Remove(pos)
        # 3. Node = Sub menu
        elif self.text!=None:
            # Remove the wxMenuItem from the parent menu
            parent = self._wx_parent()
            parent.Remove(self._id)
            # Make the ID available again
            self.root._IDmanager.freeID(self._id)
        # 3. Node = Embedded menu
#        else: pass

        MenuNode._destroy_wx(self)


    def _is_menubar(self):
        """Returns True if the node represents a menu bar."""
        return hasattr(self, "_wx_menubar")

    def _is_embedded_menu(self):
        """Returns True if the node represents an embedded menu."""
        return self.text==None and self.parent!=None and (not self._is_menubar())

    def _get_embedded_items(self):
        """
        May only be called on embedded menus.
        """

        res = []
        for cn in self._children:
            if cn._wx!=None:
                res.append(cn)
            elif isinstance(cn, Menu):
                res += cn._get_embedded_items()
        return res
        

    def _connect(self, node):
        """Connect a menu event with the handler.

        May only be called on the root node.
        """

        id = node._wx.GetId()
        wx.EVT_MENU(self._window, id, self._onMenuSelection)
        self._menu_id_lut[id] = node

    def _disconnect(self, node):
        """Disconnect a menu event from the handler.

        May only be called on the root node.
        """
        
        id = node._wx.GetId()
        print(("Disconnecting",node.name,"ID:",id))
        wx.EVT_MENU(self._window, id, None)
        del self._menu_id_lut[id]


    def _onMenuSelection(self, event):
        print(("Item",event.GetId(),"selected"))
        item = self._menu_id_lut.get(event.GetId(), None)
        if item==None:
            print ("Menu item ID not found!")
            return

        cmd = item.command
        if cmd!=None:
            cmd()
        

    def _dump(self, depth=0):
        print(("%s%s"%(2*depth*" ",self)))
        for n in self._children:
            if hasattr(n, "_dump"):
                n._dump(depth+1)
            else:
                print(("%s%s"%(2*(depth+1)*" ",n)))




# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: eventmanager.py,v 1.4 2005/03/08 16:13:42 mbaas Exp $

## \file eventmanager.py
## Contains the EventManager class.

import sys, types, bisect
from copy import copy

# Receiver
class _Receiver:
    """Stores a receiver (=callable) and a priority.

    This class defines a comparison operator (that's why this class is
    used instead of a tuple (priority, receiver)).
    """
    def __init__(self, receiver, priority=1):
        self.receiver = receiver
        self.priority = priority


    def __str__(self):
        # If the receiver is an instance method, then obtain the corresponding
        # class name
        cls = getattr(self.receiver, "im_class", None)
        if cls!=None:
            s = getattr(cls, "__name__", "?")+"."
        else:
            s = ""
        s += getattr(self.receiver, "__name__", "<unnamed>")
        return "(%d, %s)"%(self.priority, s)

    __repr__ = __str__

    def __lt__(self, other):
        return self.priority<other.priority
    

# EventManager
class EventManager:
    """Manages any kind of events.

    System wide event receivers must indicate if the event should
    be consumed (return value = True) or passed to the scene wide
    receivers.
    """

    def __init__(self):
        """Constructor."""

        # System wide connections.
        # Key: Event name - Value: Sorted list of _Receivers
        self.system_connections = {}
        # Scene wide connections.
        # Key: Event name - Value: Sorted list of _Receivers
        self.scene_connections = {}

    def __str__(self):
        s = 70*"-"+"\n"
        s += "System events\n"
        s += 70*"-"+"\n"
        for event in self.system_connections:
            s+='Event: "%s"\n'%event
            for rec in self.system_connections[event]:
                s+="  %s\n"%rec
        s+="\n"
                
        s += 70*"-"+"\n"
        s += "Scene events\n"
        s += 70*"-"+"\n"
        for event in self.scene_connections:
            s+='Event: "%s"\n'%event
            for rec in self.scene_connections[event]:
                s+="  %s\n"%rec
        return s
        

    # event
    def event(self, name, *params, **keyargs):
        """Signal an event.

        When an event handler returns True the notification is interrupted
        (i.e. the event is consumed and not available anymore for other
        handlers).

        \param name (\c str) Name of the event.
        \return True, if any event handler returned True.

        \todo Aufruf der Empfaenger-Methode kann Exception verursachen.
        """

        # Process system wide connections
        receivers = self.system_connections.get(name, [])
        for rec in copy(receivers):
            if rec.receiver(*params, **keyargs):
                return True

        # Process scene wide connections
        receivers = self.scene_connections.get(name, [])
        for rec in copy(receivers):
            if rec.receiver(*params, **keyargs):
                return True

        return False

    # connect
    def connect(self, name, receiver, priority=10, system=False):
        """Connect a function or method to an event.

        The priority determines the calling order when the corresponding
        event is emitted. Receivers with a smaller priority are invoked first.

        \param name (\c str) Name of the event.
        \param receiver The receiving function or method, or an instance
                        of a class that must implement an on<Event>() method.
        \param priority (\c int) Priority of the receiver
        \param system (\c bool) Specifies if the connection is system wide or
                      not (default: \c False).
        """

        receiver = self._determine_receiver(name, receiver)

        # Disconnect the receiver if it was already connected
        try:
            self.disconnect(name,receiver,system)
        except KeyError:
            pass

        if system:
            connections = self.system_connections
        else:
            connections = self.scene_connections

        rec = _Receiver(receiver, priority)
        # Has the event already any connections? then add the new receiver
        if name in connections:
            bisect.insort(connections[name], rec)
#            connections[name].append(receiver)
        # otherwise create a new list
        else:
            connections[name] = [rec]

        return (name,receiver)

    # disconnect
    def disconnect(self, name, receiver=None, system=False):
        """Disconnect a function or method from an event.

        \param name (\c str) Name of the event.
        \param receiver The receiving function or method, or an instance
                        of a class that must implement an on<Event>() method.
        \param system (\c bool) Specifies if the connection is system wide or
                      not (default: \c False).
        """

        if isinstance(name,tuple):
            name,receiver = name

        if system:
            connections = self.system_connections
        else:
            connections = self.scene_connections

        # Are all connections to be removed?
        if receiver==None:
            if name in connections:
                del connections[name]
                return

        receiver = self._determine_receiver(name, receiver)

        # Has the event any connections at all?
        if name not in connections:
            raise KeyError('Receiver is not connected to event "%s"'%name)

        # Try to remove the connection
#        try:
#            connections[name].remove(receiver)
#        except ValueError:
#            raise KeyError('Receiver is not connected to event "%s"'%name)
        
        for i,rec in enumerate(connections[name]):
            if rec.receiver==receiver:
                break
        else:
            raise KeyError('Receiver is not connected to event "%s"'%name)

        del connections[name][i]

    # disconnectAll
    def disconnectAll(self, system=False):
        """Remove all connections.

        \param system (\c bool) Specifies if the system wide connections or
                      the scene wide connections shall be removed
                      (default: \c False).
        """
        if system:
            self.system_connections = {}
        else:
            self.scene_connections = {}


    ## private:
    def _determine_receiver(self, name, receiver):
        """Returns the receiver object.

        The receiver is the callable that gets called when an event
        occurs.

        \param name (\c str) Name of the event.
        \param receiver The receiving function or method, or an instance
                        of a class that must implement an on<Event>() method.
        \return Callable object.
        """

        methodname = "on"+name
        if hasattr(receiver, methodname):
            return getattr(receiver, methodname)
        
        if hasattr(receiver, "__call__"):
            return receiver
            
        raise ValueError("Receiver argument must be a callable or an object with an %s() method"%methodname)
        
######################################################################

_global_event_manager = EventManager()

# eventManager
def eventManager():
    """Returns the global event manager.

    \return Event manager (EventManager)
    """
    global _global_event_manager
    return _global_event_manager



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: camerabase.py,v 1.1 2005/05/12 12:58:28 mbaas Exp $

## \file camerabase.py
## Contains the CameraBase class.

"""This module contains the CameraBase class."""

from .Interfaces import *
from . import protocols
from . import slots
from .cgtypes import *
from math import pi
from .worldobject import WorldObject
from .globalscene import getScene
from . import _core

# CameraBase
class CameraBase(WorldObject):
    """Base class for camera objects.
    """

    def __init__(self,
                 auto_nearfar = True,
                 nearplane = 0.1,
                 farplane = 1000,
                 **params):
        WorldObject.__init__(self, **params)

        self.nearplane_slot = slots.DoubleSlot(nearplane)
        self.farplane_slot = slots.DoubleSlot(farplane)
        self.autonearfar_slot = slots.BoolSlot(auto_nearfar)
        self.addSlot("nearplane", self.nearplane_slot)
        self.addSlot("farplane", self.farplane_slot)
        self.addSlot("autonearfar", self.autonearfar_slot)

    # "output" property...
    exec(slots.slotPropertyCode("nearplane"))
    exec(slots.slotPropertyCode("farplane"))
    exec(slots.slotPropertyCode("autonearfar"))
    

    def protocols(self):
        return [ISceneItem, IComponent, IWorldObject, ICamera]

    # eyeRay
    def eyeRay(self, x0, y0, width, height):
        """Return a ray from the eye position through an image point.

        This method returns a ray whose origin is at the eye position
        and that goes through a given point on the image plane. The
        point on the plane is given by (x0, y0) which each ranges from
        0 to 1. (0,0) is at the upper left and (1,1) at the lower right.
        The arguments width and height determine the ratio of the image
        plane (the absolute values of width and height are irrelevant).
        The return value is a 2-tuple (p,u) where p is the ray origin
        and u the normalized direction. Both vectors are given in world
        space.
        """
        V = self.viewTransformation()
        P = self.projection(width, height, 1, 10)
        R = mat4().rotation(pi, vec3(0,1,0))
        if getScene().handedness=='l':            
            S = mat4().scaling(vec3(-1,1,1))
            I = (P*S*R*V).inverse()
        else:
            I = (P*R*V).inverse()
        x = 2.0*x0-1.0
        y = 1.0-2.0*y0
        q = I*vec3(x,y,0)
        p = self.worldtransform[3]
        p = vec3(p.x, p.y, p.z)
        return (p, (q-p).normalize())

    # getNearFar
    def getNearFar(self):
        """Return the distances to the near and far clipping plane.

        If auto_nearfar is True, the near/far values are computed from
        the scene extent, otherwise the stored values are used.
        
        Compute near and far clipping plane distances from the bounding
        box of the scene. The scene bounding box is converted to a
        bounding sphere and the near and far clipping planes are set
        as tangent planes to the bounding sphere.
        """

        if not self.autonearfar:
            return self.nearplane, self.farplane

        # Get the bounding box of the entire scene
        bbox = getScene().boundingBox()

        # Determine bounding sphere
        bmin,bmax = bbox.getBounds()
        if bmin!=None and bmin!=bmax:
            # Box center (resp. sphere center)
            c = 0.5*(bmin+bmax)
            # Radius of the bounding sphere
            r = (bmax-c).length()
        else:
            c = vec3(0,0,0)
            r = 10

        # Transformation World->Camera
        VT = self.viewTransformation()

#        minnear = (bmax-bmin).length()/1000
        minnear = self.nearplane
        minfar = minnear+1

        # cz: Depth of the center point
        cz = (VT*c).z
        near = max(cz-r, minnear)
        far  = max(cz+r, minfar)

        if (far-near)<0.01:
            far+=1

        return (near,far)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: noise.py,v 1.3 2006/02/14 19:29:39 mbaas Exp $

"""This module contains various noise functions.
"""

from ._core import noise, snoise, cellnoise, scellnoise, fBm, turbulence
from ._core import pnoise as _pnoise
from ._core import spnoise as _spnoise
from ._core import vnoise, vsnoise, vcellnoise, vscellnoise, vfBm, vturbulence
from ._core import vpnoise as _vpnoise
from ._core import vspnoise as _vspnoise
from .cgtypes import vec3

# pnoise
def pnoise(*args):
    """pnoise(point, period) / pnoise(point, t, pperiod, tperiod) -> float

    Periodic noise function. Basically this is the same than noise
    but with a periodic return value: pnoise(point) = pnoise(point+period).
    The time value can be either part of the point or it can be
    specified separately. The point and period must always have the
    same dimension. The return value is in the range from 0 to 1.
    """
    
    n = len(args)
    if n==2:
        v, pv = args
        try:
            m = len(v)
        except:
            return _pnoise(v, 0, pv, 1)
        if m==1:
            return _pnoise(v[0], 0, pv[0], 1)
        elif m==2:
            x,y = v
            px,py = pv
            return _pnoise(x,y,px,py)
        elif m==3:
            x,y,z = v
            px,py,pz = pv
            return _pnoise(x,y,z,px,py,pz)
        elif m==4:
            x,y,z,t = v
            px,py,pz,pt = pv
            return _pnoise(x,y,z,t,px,py,pz,pt)
        else:
            raise ValueError("Invalid arguments")
        
    elif n==4:
        v,t,pv,pt = args
        try:
            m = len(v)
        except:
            return _pnoise(v, t, pv, pt)
        if m==1:
            return _pnoise(v[0], t, pv[0], pt)
        elif m==2:
            x,y = v
            px,py = pv
            return _pnoise(x,y,t,px,py,pt)
        elif m==3:
            x,y,z = v
            px,py,pz = pv
            return _pnoise(x,y,z,t,px,py,pz,pt)
        else:
            raise ValueError("Invalid arguments")
    
    else:
        raise TypeError("only 2 or 4 arguments allowed")


# spnoise
def spnoise(*args):
    """spnoise(point, period) / spnoise(point, t, pperiod, tperiod) -> float

    Signed periodic noise function. The return value is in the range
    from -1 to 1. A call to spnoise(args) is equivalent to
    2*pnoise(args)-1.
    """
    
    n = len(args)
    if n==2:
        v, pv = args
        try:
            m = len(v)
        except:
            return _spnoise(v, 0, pv, 1)
        if m==1:
            return _spnoise(v[0], 0, pv[0], 1)
        elif m==2:
            x,y = v
            px,py = pv
            return _spnoise(x,y,px,py)
        elif m==3:
            x,y,z = v
            px,py,pz = pv
            return _spnoise(x,y,z,px,py,pz)
        elif m==4:
            x,y,z,t = v
            px,py,pz,pt = pv
            return _spnoise(x,y,z,t,px,py,pz,pt)
        else:
            raise ValueError("Invalid arguments")
        
    elif n==4:
        v,t,pv,pt = args
        try:
            m = len(v)
        except:
            return _spnoise(v, t, pv, pt)
        if m==1:
            return _spnoise(v[0], t, pv[0], pt)
        elif m==2:
            x,y = v
            px,py = pv
            return _spnoise(x,y,t,px,py,pt)
        elif m==3:
            x,y,z = v
            px,py,pz = pv
            return _spnoise(x,y,z,t,px,py,pz,pt)
        else:
            raise ValueError("Invalid arguments")
    
    else:
        raise TypeError("only 2 or 4 arguments allowed")


# vpnoise
def vpnoise(*args):
    """vpnoise(point, period) / vpnoise(point, t, pperiod, tperiod) -> noiseval

    Periodic noise function. Basically this is the same than vnoise
    but with a periodic return value: vpnoise(point) = vpnoise(point+period).
    The time value can be either part of the point or it can be
    specified separately. The point and period must always have the
    same dimension. The components of the return value are in the range
    from 0 to 1.
    """
    
    n = len(args)
    if n==2:
        v, pv = args
        try:
            m = len(v)
        except:
            return _vpnoise(v, pv)
        if m==1:
            return _vpnoise(v[0], pv[0])
        elif m==2:
            x,y = v
            px,py = pv
            return _vpnoise(x,y,px,py)
        elif m==3:
            x,y,z = v
            px,py,pz = pv
            return _vpnoise(x,y,z,px,py,pz)
        elif m==4:
            x,y,z,t = v
            px,py,pz,pt = pv
            return _vpnoise(x,y,z,t,px,py,pz,pt)
        else:
            raise ValueError("Invalid arguments")
        
    elif n==4:
        v,t,pv,pt = args
        try:
            m = len(v)
        except:
            return _vpnoise(v, t, pv, pt)
        if m==1:
            return _vpnoise(v[0], t, pv[0], pt)
        elif m==2:
            x,y = v
            px,py = pv
            return _vpnoise(x,y,t,px,py,pt)
        elif m==3:
            x,y,z = v
            px,py,pz = pv
            res = _vpnoise(x,y,z,t,px,py,pz,pt)
            return vec3(res.x, res.y, res.z)
        else:
            raise ValueError("Invalid arguments")
    
    else:
        raise TypeError("only 2 or 4 arguments allowed")


# vspnoise
def vspnoise(*args):
    """vspnoise(point, period) / vspnoise(point, t, pperiod, tperiod) -> noiseval

    Periodic noise function. Basically this is the same than vsnoise
    but with a periodic return value: vspnoise(point) = vspnoise(point+period).
    The time value can be either part of the point or it can be
    specified separately. The point and period must always have the
    same dimension. The components of the return value are in the range
    from -1 to 1.
    """
    
    n = len(args)
    if n==2:
        v, pv = args
        try:
            m = len(v)
        except:
            return _vspnoise(v, pv)
        if m==1:
            return _vspnoise(v[0], pv[0])
        elif m==2:
            x,y = v
            px,py = pv
            return _vspnoise(x,y,px,py)
        elif m==3:
            x,y,z = v
            px,py,pz = pv
            return _vspnoise(x,y,z,px,py,pz)
        elif m==4:
            x,y,z,t = v
            px,py,pz,pt = pv
            return _vspnoise(x,y,z,t,px,py,pz,pt)
        else:
            raise ValueError("Invalid arguments")
        
    elif n==4:
        v,t,pv,pt = args
        try:
            m = len(v)
        except:
            return _vspnoise(v, t, pv, pt)
        if m==1:
            return _vspnoise(v[0], t, pv[0], pt)
        elif m==2:
            x,y = v
            px,py = pv
            return _vspnoise(x,y,t,px,py,pt)
        elif m==3:
            x,y,z = v
            px,py,pz = pv
            res = _vspnoise(x,y,z,t,px,py,pz,pt)
            return vec3(res.x, res.y, res.z)
        else:
            raise ValueError("Invalid arguments")
    
    else:
        raise TypeError("only 2 or 4 arguments allowed")





# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2010
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import sys
import fractions
from .ffmpeg import decls, cppdefs, avutil, swscale, avformat, avcodec
import ctypes

try:
    import Image
    _pilImportException = None
except ImportError as exc:
    _pilImportException = exc
    
try:
    import numpy
    _numpyImportException = None
except ImportError as exc:
    _numpyImportException = exc

# Pixel formats
RGB = 0
BGR = 1
RGBA = 2
ARGB = 3
BGRA = 4
ABGR = 5
GRAY = 6
 
# Pixel access
WIDTH_HEIGHT = 100
HEIGHT_WIDTH = 101

# Color access
SEPARATE_CHANNELS = 200
COMBINED_CHANNELS = 201

class MediaFileError(Exception):
    pass

class VideoData:
    """Decoded image data.
    
    This class serves as a container for the data that is passed to the
    video callbacks.
    """
    def __init__(self, stream, size, srcPixFmt, frame=None, pts=None):
        """Constructor.
        """
        # The parent stream object (VideoStream_ffmpeg)
        self.stream = stream
        # Presentation timestamp (float)
        self.pts = pts
        # Image size (width, height)
        self.size = size
        
        # The source pixel format of the video stream.
        # This is one of the ffmpeg PIX_FMT_* integer values and it's used to
        # convert the video frame into the target format.
        self._srcPixFmt = srcPixFmt
        
        # AVFrame object (containing the decoded frame (but not as RGB))
        # The contents of this object change on every frame iteration, so the
        # same object is used for every frame.
        self._frame = frame
        
        # If the frame is queried as a numpy array, this will be the array that
        # is kept between calls.
        self._numpyArray = None
        # Stores a tuple (pixelFormat, pixelAccess, colorAccess) that contains
        # the values that were used to initialize the numpy array. Whenever
        # one of these values changes, the array needs to be re-initialized.
        self._numpyBufferParams = None
        # The sws context that is used to convert the image into a RGB numpy array
        self._numpySwsCtx = None
        # The uint*[4] array containing the data pointers into the numpy array
        # (actually only, the first item is used, the others are NULL). This
        # is passed into the sws_scale() function to fill the numpy buffer.
        self._dataPtrs = None
        # The int[4] array of line sizes (only the first one is used). This
        # is passed into the sws_scale() function to fill the numpy buffer.
        self._lineSizes = None
        
        # AVPicture object that contains the target buffer for conversion to PIL.
        # This is just a buffer that can hold the converted image (just like
        # a numpy array).
        self._picture = None
        # The size in bytes of a full (RGB) picture
        self._pictureSize = None
        # The sws context for conversion into a PIL image
        self._pilSwsCtx = None
    
    def _free(self):
        """Free any resources that have been allocated by the object.
        
        This must be called by the stream after all frames have been processed.
        """
        if self._numpySwsCtx is not None:
            swscale.sws_freeContext(self._numpySwsCtx)
            self._numpySwsCtx = None
            
        if self._pilSwsCtx is not None:
            swscale.sws_freeContext(self._pilSwsCtx)
            self._pilSwsCtx = None
        
        if self._picture is not None:
            avcodec.avpicture_free(self._picture)
            self._picture = None

        
    def isKeyFrame(self):
        """Check whether the current frame is a key frame or not.
        """
        return self._frame.key_frame==1
    
    def numpyArray(self, pixelFormat=RGB, pixelAccess=WIDTH_HEIGHT, colorAccess=SEPARATE_CHANNELS):
        """Return the current frame as a numpy array.

        pixelFormat specifies the number and order of the channels in the
        returned buffer. The source video frames are converted into the
        specified format.
        
        pixelAccess defines the order of the x,y indices when accessing
        a pixel in the returned buffer. The value can either be WIDTH_HEIGHT
        to specify the pixel position as (x,y) or HEIGHT_WIDTH if (y,x)
        is required.
        
        colorAccess specifies whether the color channels should be accessed
        as a third index (SEPARATE_CHANNELS) or if the entire pixel should
        be stored as a single integer value (COMBINED_CHANNELS). The latter
        is only allowed if the pixel format contains 1 or 4 channels.
        
        Note that the pixelAccess and colorAccess parameters do not affect
        the underlying memory layout of the buffer, they only affect the way
        that individual pixels are accessed via the numpy array.
        The memory layout is always so that rows are stored one after another
        without any gaps and the channel values of an individual pixel are
        always stored next to each other.  
        
        The underlying numpy array is reused for every frame, so if you need
        to keep the array you have to copy it.
        The returned array has a shape of (width,height,3) and a dtype of uint8.
        """
        # Make sure numpy has been imported successfully
        global _numpyImportException
        if _numpyImportException is not None:
            raise _numpyImportException
        
        # Do we need to allocate a new numpy buffer?
        if self._numpyArray is None or (pixelFormat, pixelAccess, colorAccess)!=self._numpyBufferParams:
            self._initNumpyProcessing(pixelFormat, pixelAccess, colorAccess)

        # Convert the frame into an rgb image. The result is put directly into
        # the numpy array.
        height = self.size[1]
        swscale.sws_scale(self._numpySwsCtx, self._frame.data, self._frame.linesize, 0, height, self._dataPtrs, self._lineSizes)
        
        return self._numpyArray
    
    def _initNumpyProcessing(self, pixelFormat, pixelAccess, colorAccess):
        """Initialize decoding into a numpy array.
        
        This has to be called at least once before a frame is converted into
        a numpy array.
        If the layout of the numpy array changes, the method has to be called
        again.
        
        The method creates the numpy array, the swscale context to convert
        the image data and the _lineSizes and _dataPtrs arrays for the sws_scale()
        call.
        """
        # Check the pixelAccess value...
        if pixelAccess not in [WIDTH_HEIGHT, HEIGHT_WIDTH]:
            raise ValueError("Invalid pixelAcess value")
        
        # Check the colorAccess value...
        if colorAccess not in [SEPARATE_CHANNELS, COMBINED_CHANNELS]:
            raise ValueError("Invalid colorAccess value")
        
        # Check the pixel format...
        numChannels = None
        dstPixFmt = None
        if pixelFormat==RGB:
            numChannels = 3
            dstPixFmt = decls.PIX_FMT_RGB24
        elif pixelFormat==BGR:
            numChannels = 3
            dstPixFmt = decls.PIX_FMT_BGR24
        elif pixelFormat==RGBA:
            numChannels = 4
            dstPixFmt = decls.PIX_FMT_RGBA
        elif pixelFormat==ARGB:
            numChannels = 4
            dstPixFmt = decls.PIX_FMT_ARGB
        elif pixelFormat==ABGR:
            numChannels = 4
            dstPixFmt = decls.PIX_FMT_ABGR
        elif pixelFormat==BGRA:
            numChannels = 4
            dstPixFmt = decls.PIX_FMT_BGRA
        elif pixelFormat==GRAY:
            numChannels = 1
            dstPixFmt = decls.PIX_FMT_GRAY8
        else:
            raise ValueError("Invalid pixelFormat value")

        width,height = self.size
        # Allocate the numpy array that will hold the converted image.
        # The memory layout of the buffer is always so that the image is stored
        # in rows and all the channels are stored together with a pixel.
        # Example: Row 0: RGB-RGB-RGB-RGB...
        #          Row 1: RGB-RGB-RGB-RGB...
        #          ...
        # To get from one channel to the next channel, you always have to add 1 byte
        # (for int8 channels). To get to the next x position, you have to add
        # numChannels bytes and to get to the next y position you have to add
        # width*numChannels bytes.
        # The pixelAccess and colorAccess parameters don't affect this memory
        # layout, they only affect how a pixel is accessed via numpy.
        if colorAccess==SEPARATE_CHANNELS:
            # Allocate a RGB buffer...
            if pixelAccess==WIDTH_HEIGHT:
                self._numpyArray = numpy.empty((width,height,numChannels), dtype=numpy.uint8)
                # Adjust the strides so that image rows are consecutive
                self._numpyArray.strides = (numChannels, numChannels*width, 1)
            else:
                self._numpyArray = numpy.empty((height,width,numChannels), dtype=numpy.uint8)
        else:
            if numChannels not in [1,4]:
                raise ValueError("COMBINED_CHANNELS pixel access can only be used with 1-channel or 4-channel pixel formats")
            
            if numChannels==1:
                dtype = numpy.uint8
            else:
                dtype = numpy.uint32
            
            # Allocate a RGBA buffer (where a RGBA value is stored as a uint32)
            if pixelAccess==WIDTH_HEIGHT:
                self._numpyArray = numpy.empty((width,height), dtype=dtype, order="F")
            else:
                self._numpyArray = numpy.empty((height,width), dtype=dtype)

        # Free any previously allocated context 
        if self._numpySwsCtx is not None:
            swscale.sws_freeContext(self._numpySwsCtx)
            self._numpySwsCtx = None

        # Allocate the swscale context
        self._numpySwsCtx = swscale.sws_getContext(width, height, self._srcPixFmt, 
                                                   width, height, dstPixFmt, 1)

        # Initialize the lineSizes and dataPtrs array that are passed into sws_scale().
        # The first data pointer points to the beginning of the numpy buffer.
        DataPtrType = ctypes.POINTER(ctypes.c_uint8)
        self._lineSizes = (4*ctypes.c_int)(numChannels*width,0,0,0)
        self._dataPtrs = (4*DataPtrType)(self._numpyArray.ctypes.data_as(DataPtrType), None, None, None)
        
        # Store the parameters so that we can detect parameter changes
        self._numpyBufferParams = (pixelFormat, pixelAccess, colorAccess)
    
    def pilImage(self):
        """Return the current frame as a PIL image.
        """
        global _pilImportException
        if _pilImportException is not None:
            raise _pilImportException

        # Do we need to initialize the PIL conversion?
        if self._pilSwsCtx is None:
            self._initPilProcessing()

        # Convert the image into RGB format (the convered image is stored
        # in the AVPicture buffer)...
        height = self.size[1]
        swscale.sws_scale(self._pilSwsCtx, self._frame.data, self._frame.linesize, 0, height, self._picture.data, self._picture.linesize)
            
        # Obtain a Python string containing the RGB image data...
        dataStr = ctypes.string_at(self._picture.data[0], self._pictureSize)
        
        # Convert to PIL image
        return Image.fromstring("RGB", self.size, dataStr)

    def _initPilProcessing(self):
        """Initialize decoding into a PIL image.
        
        Initializes an intermediate buffer (AVPicture) that will receive
        the converted image and a swscale context to do the conversion. 
        """
        width,height = self.size
        
        dstPixFmt = decls.PIX_FMT_RGB24
        
        # In case, this is called several times, make sure we don't leak memory.
        self._free()
        
        # Allocate a buffer that can hold the converted image
        self._picture = decls.AVPicture()
        avcodec.avpicture_alloc(self._picture, dstPixFmt, width, height)
        # Get the size of the destination image buffer
        self._pictureSize = avcodec.avpicture_get_size(dstPixFmt, width, height)
        
        # Allocate the swscale context
        self._pilSwsCtx = swscale.sws_getContext(width, height, self._srcPixFmt, 
                                                 width, height, dstPixFmt, 1)


class AudioData:
    """Decoded audio data.
    """
    def __init__(self, stream, pts=None, channels=None, framerate=None, samples=None, sampleSize=None):
        # The parent stream
        self.stream = stream
        # Presentation timestamp as an integer in stream time base units
        self.pts = pts
        # The number of channels (int)
        self.channels = channels
        # The framerate in Hz
        self.framerate = framerate
        # The decoded samples (this is a ctypes array of shorts)
        self.samples = samples
        # The number of bytes in the samples buffer
        self.sampleSize = sampleSize


class StreamBase_ffmpeg(object):
    """Base class for the audio/video streams.
    """
    def __init__(self, stream):
        """Constructor.
        
        stream is an AVStream object.
        """
        object.__init__(self)

        # Data callback
        self._dataCallback = None
        
        # AVStream object
        self._stream = stream

        # Get the codec context of the stream (AVCodecContext)
        self._codecCtx = stream.codec.contents

        # Obtain an appropriate decoder (AVCodec)...
        self._codec = avcodec.avcodec_find_decoder(self._codecCtx.codec_id)

        if self._codec is not None:
            # Open the codec
            avcodec.avcodec_open(self._codecCtx, self._codec)

    def close(self):
        """Close the stream/codec.
        """
        if self._codecCtx is not None:
            avcodec.avcodec_close(self._codecCtx)
            self._codecCtx = None
            self._codec = None
    
    def setDataCallback(self, callback):
        """Set a callback that gets called with decoded data.
        
        *callback* is a callable that must take a data object as input.
        It may also be ``None`` to remove any previously set callback.
        """
        self._dataCallback = callback
    
    def getDataCallback(self):
        """Return the currently set data callback.
        """
        return self._dataCallback
    
    def decodeBegin(self):
        """This is called right before decoding begins.
        
        The method returns a boolean indicating whether the stream should
        be enabled (True) or not (False). A stream might be enabled if there
        is no callback (i.e. noone is interested in the decoded data).
        """
        pass
    
    def decodeEnd(self):
        """This is called after decoding has finished.
        
        This is only called when decodeBegin() didn't return False or raise
        an error.
        """
        pass
    
    def handlePacket(self, pkt):
        """This is called during decoding to handle an encoded file packet.
        
        This is only called when decodeBegin() didn't return False or raise
        an error.
        Yields the decoded data (in case of audio data, a packet may contain
        more than one frame).
        """
        pass

    @property
    def index(self):
        """Return the stream index.
        """
        return self._stream.index
        
    @property
    def fourCC(self):
        v = self._codecCtx.codec_tag
        return "%s%s%s%s"%(chr(v&0xff), chr((v>>8)&0xff), chr((v>>16)&0xff), chr((v>>24)&0xff))

    @property
    def codecName(self):
        if self._codec is None:
            return None
        else:
            return self._codec.name

    @property
    def codecLongName(self):
        if self._codec is None:
            return None
        else:
            return self._codec.long_name

    @property
    def bitRate(self):
        """Return the average bit rate as an integer (bits per second).
        """
        return self._codecCtx.bit_rate

    @property
    def timeBase(self):
        """Return the time base as a Fraction object.
        
        This is the fundamental unit of time (in seconds) in terms of which
        frame timestamps are represented
        """
        tb = self._stream.time_base
        num = tb.num
        den = tb.den
        # Make sure we don't divide by 0
        if den==0:
            num = 0
            den = 1
        return fractions.Fraction(num, den)
    
    @property
    def duration(self):
        """Return the duration in timeBase units.
        """
        return self._stream.duration
    


class AudioStream_ffmpeg(StreamBase_ffmpeg):
    """Audio stream class.
    """
    def __init__(self, stream):
        StreamBase_ffmpeg.__init__(self, stream)
    
    def decodeBegin(self):

        # A dummy packet that is used to pass the output buffer to avcodec_decode_audio3
        self._dummyPkt = decls.AVPacket()
        avcodec.av_init_packet(self._dummyPkt)

        # Allocate a buffer for the decoded audio frame
        size = 192000
        self._sampleBuf = (size*ctypes.c_short)()

        self._audioData = AudioData(self, channels=self._codecCtx.channels, framerate=self._codecCtx.sample_rate, samples=self._sampleBuf)
        return True

    def decodeEnd(self):
        pass
    
    def handlePacket(self, pkt):
        # Decode the frame...
        codecCtx = self._codecCtx
        
        # Set the data pointer and size of the dummy packet to that of packet
        # (the dummy packet is used because there may be more than one frame
        # and we have to adjust the data pointer and size to decode all of them)
        dummyPkt = self._dummyPkt
        dummyPkt.data = pkt.data
        dummyPkt.size = pkt.size
        
        bytesUsed = 1
        while bytesUsed>0 and bool(dummyPkt.data):
            # Decode the next frame
            frameSize,bytesUsed = avcodec.avcodec_decode_audio3(codecCtx, self._sampleBuf, dummyPkt)
            # frameSize is the number of bytes that have been written into the sample buffer
            # bytesUsed is the number of bytes that have been used in the input buffer

            # The following two lines are equivalent to dummyPkt.data += bytesUsed
            # (the first line converts the pointer value to an int, the second
            # line adds bytesUsed and converts the result back to a ctypes pointer)
            addr = ctypes.addressof(dummyPkt.data.contents)
            dummyPkt.data = ctypes.cast(addr+bytesUsed, type(dummyPkt.data))
            dummyPkt.size -= bytesUsed
            
            # Report the frame if we have one
            if frameSize>0:
                self._audioData.pts = pkt.pts
                self._audioData.sampleSize = frameSize
                
                yield self._audioData

    @property
    def numChannels(self):
        return self._codecCtx.channels

    @property
    def sampleRate(self):
        """Return the sample rate as an integer (samples per second).
        """
        return self._codecCtx.sample_rate


class VideoStream_ffmpeg(StreamBase_ffmpeg):
    """Video stream class.
    """

    def __init__(self, stream):
        """Constructor.
        
        stream is an AVStream object containing the video stream.
        """
        StreamBase_ffmpeg.__init__(self, stream)
        self._frame = None
        self._videoData = None
    
    def decodeBegin(self):
        
        # Allocate a video frame that will store the decoded frames and
        # create the VideoData object that will be passed to the callers
        # (the actual conversion to RGB (or whatever) is done by the
        # VideoData object)
        try:
            self._frame = avcodec.avcodec_alloc_frame()
            if self._frame is None:
                raise MemoryError("Failed to allocate AVFrame object")
            
            width = self._codecCtx.width
            height = self._codecCtx.height
            self._videoData = VideoData(self, size=(width, height), frame=self._frame, srcPixFmt=self._codecCtx.pix_fmt)
            
            tb = self._stream.time_base
            self._timebase = float(tb.num)/tb.den
        except:
            self.decodeEnd()
            raise

        return True

    def decodeEnd(self):
        if self._videoData is not None:
            self._videoData._free()
            self._videoData = None
        if self._frame is not None:
            avutil.av_free(self._frame)
            self._frame = None

    def handlePacket(self, pkt):
        # Decode the frame...
        codecCtx = self._codecCtx
        hasFrame,bytes = avcodec.avcodec_decode_video2(codecCtx, self._frame, pkt)

        if hasFrame:
            # Conversion into RGB or creation of a PIL image is done by the
            # video data object on demand
            
            #print "PTS:%s  DTS:%s  Duration:%s"%(pkt.pts, pkt.dts, pkt.duration)
            self._videoData.pts = pkt.dts*self._timebase
        
            yield self._videoData

    @property
    def size(self):
        """Return the width and height of a video frame in pixels.
        """
        return (self._codecCtx.width, self._codecCtx.height)
        
    @property
    def width(self):
        """Return the width of a video frame in pixels.
        """
        return self._codecCtx.width

    @property
    def height(self):
        """Return the height of a video frame in pixels.
        """
        return self._codecCtx.height
    
    @property
    def frameRate(self):
        """Return the frame rate as a Fraction object.
        """
        fr = self._stream.r_frame_rate
        num = fr.num
        den = fr.den
        # Make sure we don't divide by 0
        if den==0:
            num = 0
            den = 1
        return fractions.Fraction(num, den)
    
    @property
    def numFrames(self):
        """Return the number of frames if known.
        
        Returns an integer or None if the number of frames is not known.
        """
        n = self._stream.nb_frames
        if n==0:
            n = None
        return n
        
    @property
    def pixelAspect(self):
        """Return the pixel aspect ratio if the value is known.
        
        Returns a Fraction object or None if the aspect ratio is not known.
        """
        a = self._stream.sample_aspect_ratio
        num = a.num
        den = a.den
        if den!=0 and num!=0:
            return fractions.Fraction(num, den)
        else:
            return None


class Media_Read_ffmpeg(object):
    """Media file reader.
    """
    
    def __init__(self, fileName):
        """Constructor.
        
        fileName is the name of the media file.
        """
        object.__init__(self)
        
        # Audio streams (AudioStream_ffmpeg objects)
        self.audioStreams = []
        # Video streams (VideoStream_ffmpeg objects)
        self.videoStreams = []
    
        # All streams in file order
        self._streams = []
    
        # The AVFormatContext object for the open file
        self._formatCtx = None
        
        # Open the video file
        self._formatCtx = avformat.av_open_input_file(fileName, None, 0, None)
        # Fill the 'streams' fields...
        avformat.av_find_stream_info(self._formatCtx)

        # Create the stream wrapper objects
        for i in range(self._formatCtx.nb_streams):
            stream = self._formatCtx.streams[i].contents
            codec = stream.codec.contents
            if codec.codec_type==decls.CODEC_TYPE_VIDEO:
                stream = VideoStream_ffmpeg(stream)
                self._streams.append(stream)
                self.videoStreams.append(stream)
            elif codec.codec_type==decls.CODEC_TYPE_AUDIO:
                stream = AudioStream_ffmpeg(stream)
                self._streams.append(stream)
                self.audioStreams.append(stream)
            else:
                self._streams.append(None)

    def __enter__(self):
        return self
    
    def __exit__(self, errorType, errorValue, traceback):
        self.close()
        return False
    
    def close(self):
        """Close the file.
        """
        for stream in self.audioStreams:
            stream.close()
        for stream in self.videoStreams:
            stream.close()
        
        self.audioStreams = []
        self.videoStreams = []
        
        if self._formatCtx is not None:
            avformat.av_close_input_file(self._formatCtx)
            self._formatCtx = None
    
    def numAudioStreams(self):
        """Return the number of audio streams defined in the file.
        """
        return len(self.audioStreams)
    
    def numVideoStreams(self):
        """Return the number of video streams defined in the file.
        """
        return len(self.videoStreams)

    def iterData(self, streams=None):
        """Iterate over the frames of the first video stream.
        """
        # Use the first video stream by default...
        if streams is None:
            if len(self.videoStreams)==0:
                return
            videoStream = self.videoStreams[0]
            streams = [videoStream]

        if len(streams)==0:
            return

        # Initialize the streams...
        # streamDict: Key:Stream index - Value:Stream object
        streamDict = {}
        for stream in streams:
            # Check the type of the stream
            if not isinstance(stream, StreamBase_ffmpeg):
                raise TypeError("Stream object expected")
            # Is this a stream from someone else?
            if stream not in self._streams:
                raise ValueError("Invalid stream (stream is from a different source)")
            
            stream.decodeBegin()
            streamDict[stream.index] = stream

        try:
            # Iterate over the packets in the file and decode the frames...
            for pkt in self.iterPackets():
                stream = streamDict.get(pkt.stream_index, None)
                if stream is not None:
                    for data in stream.handlePacket(pkt):
                        yield data
            
            # Some codecs have a delay between input and output, so go on
            # feeding empty packets until we receive no more frames.            
            pkt = decls.AVPacket()
            avcodec.av_init_packet(pkt)
            for stream in list(streamDict.values()):
                hasData = True
                while hasData: 
                    hasData = False
                    for data in stream.handlePacket(pkt):
                        hasData = True
                        yield data
        finally:
            for stream in list(streamDict.values()):
                stream.decodeEnd()
    
    def decode(self):
        """Decode stream data and pass it to the callbacks stored in the streams.
        """
        # Determine the streams that have a callback
        streams = []
        for stream in self._streams:
            cb = stream.getDataCallback()
            if cb is not None:
                streams.append(stream)

        # Iterate over the data...
        for data in self.iterData(streams):
            callback = data.stream.getDataCallback()
            callback(data)

    def iterPackets(self):
        """Iterate over all raw packets in the file.
        
        Yields AVPacket objects with the current packet. Actually, it always
        yields the same object that just has a new packet in it, so it's not
        valid to store the packet for later use.
        """
        formatCtx = self._formatCtx
        if formatCtx is None:
            return

        pkt = decls.AVPacket()
        eof = False
        while not eof:
            # Read the next frame packet...
            try:
                eof = not avformat.av_read_frame(formatCtx, pkt)
                if not eof:
                    yield pkt
            finally:
                avcodec.av_free_packet(pkt)

    @property
    def title(self):
        if self._formatCtx is None:
            return ""
        else:
            return self._formatCtx.title

    @property
    def author(self):
        if self._formatCtx is None:
            return ""
        else:
            return self._formatCtx.author

    @property
    def copyright(self):
        if self._formatCtx is None:
            return ""
        else:
            return self._formatCtx.copyright

    @property
    def album(self):
        if self._formatCtx is None:
            return ""
        else:
            return self._formatCtx.album

    @property
    def comment(self):
        if self._formatCtx is None:
            return ""
        else:
            return self._formatCtx.comment

    @property
    def year(self):
        if self._formatCtx is None:
            return 0
        else:
            return self._formatCtx.year

    @property
    def track(self):
        if self._formatCtx is None:
            return 0
        else:
            return self._formatCtx.track

    @property
    def genre(self):
        if self._formatCtx is None:
            return ""
        else:
            return self._formatCtx.genre



class VideoStream_Write_ffmpeg:
    def __init__(self, formatCtx, size, frameRate, pixelAspect):
        """Constructor.
        
        formatCtx is a AVFormatContext object.
        size is a tuple (width, height) containing the resolution of the
        video images.
        frameRate is the target framerate. It can be a single int or float
        or a tuple (num,den).
        pixelAspect is a float containing the pixel aspect ratio.
        """
        self._formatCtx = formatCtx
        self._size = size
        self._pixelAspect = pixelAspect
        
        try:
            # Check if frameRate is a 2-tuple
            num,den = frameRate
        except:
            # Turn the framerate into a rational...
            r = avutil.av_d2q(frameRate, 255)
            frameRate = (r.num, r.den)
        
        # The framerate as a tuple (num, den).
        self._frameRate = frameRate

        # Create an AVStream
        self._stream = self._createStream()
        
        # Get the codec context of the stream
        self._codecCtx = self._stream.codec.contents
        
        width,height = self._size
        # Set up the buffer that can store the encoded frame
        self._bufSize = 6*width*height+200
        self._buffer = ctypes.create_string_buffer(self._bufSize)

        self._frame = avcodec.avcodec_alloc_frame()
        if self._frame is None:
            raise MemoryError("Failed to allocate AVFrame object")
        avcodec.avpicture_alloc(self._frame, self._codecCtx.pix_fmt, width, height)

        # Allocate a picture for the converted image...
        self._picture = decls.AVPicture()
        avcodec.avpicture_alloc(self._picture, self._codecCtx.pix_fmt, width, height)
        # Get the size of the destination image buffer
        self._pictureSize = avcodec.avpicture_get_size(self._codecCtx.pix_fmt, width, height)
        
        self._pkt = decls.AVPacket()
        avformat.av_init_packet(self._pkt)
        #pkt.stream_index = outputstream.index
        
        self._currentPts = 0


    def close(self):
        pass
    
    def writeFrame(self, frame):
        buf = ctypes.addressof(self._buffer)
        self._frame.pts = self._currentPts
        self._currentPts += 1
        bytes = avcodec.avcodec_encode_video(self._codecCtx, buf, self._bufSize, self._frame)
#        print "pts", self._frame.pts
        print(("bytes %s"%bytes))
        if bytes>0:
            self._pkt.data = ctypes.cast(self._buffer, ctypes.POINTER(ctypes.c_uint8))
            self._pkt.size = bytes
            self._pkt.pts = self._frame.pts
#            if self._codecCtx.coded_frame.contents.key_frame:
#                self._pkt.flags |= cppdefs.PKT_FLAG_KEY
                
            avformat.av_interleaved_write_frame(self._formatCtx, self._pkt)

    def _createStream(self):
        """Create the AVStream and open the codec.
        """
        
        # Allocate a new stream
        stream = avformat.av_new_stream(self._formatCtx, self._formatCtx.nb_streams)
        # Initialize it as a video stream
        avcodec.avcodec_get_context_defaults2(stream.codec.contents, decls.CODEC_TYPE_VIDEO)

        outputFormat = self._formatCtx.oformat.contents
        codecCtx = stream.codec.contents
        
        if outputFormat.flags & decls.AVFMT_GLOBALHEADER:
            stream.codec.contents.flags |= decls.CODEC_FLAG_GLOBAL_HEADER
        
        # Prepare the codec...
        codecId = avformat.av_guess_codec(outputFormat, decls.CODEC_TYPE_VIDEO, fileName=self._formatCtx.filename)
        if codecId is None:
            raise ValueError("Could not determine video codec to use for encoding the video.")
        codec = avcodec.avcodec_find_encoder(codecId)
        if codec is None:
            raise ValueError("Could not find encoder.")

        if codec.supported_framerates:
            print ("mediafile: supported framerates available")
        else:
            print ("mediafile: supported framerates not available")

        codecCtx.codec_id = codecId
        num,den = self._frameRate
        codecCtx.time_base.den = num
        codecCtx.time_base.num = den
        width,height = self._size
        codecCtx.width = width
        codecCtx.height = height
        codecCtx.sample_aspect_ratio = avutil.av_d2q(self._pixelAspect, 255)
        codecCtx.pix_fmt = decls.PIX_FMT_YUV420P
        stream.sample_aspect_ratio = codecCtx.sample_aspect_ratio
        
        # Check if the codec supports the pixel format (if not, switch the format)
        if codec.pix_fmts:
            i = 0
            while codec.pix_fmts[i]!=-1:
                if codecCtx.pix_fmt==codec.pix_fmts[i]:
                    break
                i += 1
            else:
                codecCtx.pix_fmt = codec.pix_fmts[0]
        
        # Open the codec
        avcodec.avcodec_open(codecCtx, codec)
        
        return stream
    

class Media_Write_ffmpeg(object):
    """Media file writer.
    """
    
    def __init__(self, fileName):
        object.__init__(self)
        
        self._formatCtx = None
        self._headerWritten = False
        
        # A list of VideoStream_Write_ffmpeg objects
        self._videoStreams = []
        
        # Create and initialize the AVFormatContext structure...
        self._formatCtx = self._createFormatContext(fileName)
        
        # Open the file
        status = avformat.url_fopen(self._formatCtx.pb, fileName, decls.URL_WRONLY)
        if status<0:
            self._freeFormatContext()
            raise MediaFileError('Could not write file "%s"'%fileName)

    def close(self):
        """Close the file.
        """
        if self._formatCtx is not None:
            for stream in self._videoStreams:
                stream.close()
            self._videoStreams = []

            if self._formatCtx.pb:
                if self._headerWritten:
                    avformat.av_write_trailer(self._formatCtx)
                avformat.url_fclose(self._formatCtx.pb)
            self._freeFormatContext()

    def createVideoStream(self, size, frameRate=30, pixelAspect=1.0):
        """Add a new video stream to the file.
        """
        try:
            stream = VideoStream_Write_ffmpeg(self._formatCtx,
                                              size=size,
                                              frameRate=frameRate,
                                              pixelAspect=pixelAspect)
            self._videoStreams.append(stream)
        except:
            self.close()
            raise

    def writeFrame(self, frame):
        if not self._headerWritten:
            self._writeHeader()
        
        self._videoStreams[0].writeFrame(frame)
    
    def _writeHeader(self):
        avformat.av_write_header(self._formatCtx)
        self._headerWritten = True
    
    def _createFormatContext(self, fileName):
        """Create a new AVFormatContext and initialize it.
        
        fileName is the name of the output file (this will determine the format
        of the file).
        Returns a AVFormatContext object.
        """
        # Determine the output format (AVOutputFormat object) from the file name...
        fmt = avformat.av_guess_format(fileName=fileName)
        if fmt is None:
            raise MediaFileError('Cannot determine output format for file "%s"'%fileName)
        
        # Create a new AVFormatContext struct
        formatCtx = avformat.avformat_alloc_context()
        
        # Initialize some fields...
        formatCtx.oformat = ctypes.pointer(fmt)
        formatCtx.filename = fileName
        formatCtx.timestamp = 0          # Where should this timestamp come from?

        mux_preload = 0.5
        mux_max_delay = 0.7
        formatCtx.preload = int(mux_preload*cppdefs.AV_TIME_BASE)
        formatCtx.max_delay= (int)(mux_max_delay*cppdefs.AV_TIME_BASE)
        formatCtx.loop_output = cppdefs.AVFMT_NOOUTPUTLOOP
        formatCtx.flags |= cppdefs.AVFMT_FLAG_NONBLOCK
        
        # Do we have to call av_set_parameters() here? (this is done in ffmpeg.c)
        
        return formatCtx
    
    def _freeFormatContext(self):
        """Free the internal AVFormatContext struct.
        """
        if self._formatCtx is not None:
            avutil.av_free(self._formatCtx)
            self._formatCtx = None

    def writeFrame__dummy(self):
        # Create a packet struct and initialize it
        pkt = AVPacket()
        avcodec.av_init_packet(pkt)
        #pkt.stream_index = outputstream.index
        
        #big_picture.pts= ost->sync_opts;
        
        # Encode the frame
        bytes = avcodec.avcodec_encode_video(codecCtx, buf, bufsize, picture)
        if bytes>0:
            pkt.data = buf
            pkt.size = bytes
#            pkt.pts = ...
            if codecCtx.coded_frame.key_frame:
                pkt.flags |= PKT_FLAG_KEY
                
            avformat.av_interleaved_write_frame(formatCtx, pkt)


######################################################################

_av_is_registered = False

def open(name, mode="r", **keyargs):
    """Open an audio/video file.

    *name* is the name of the media file to read or write. *mode* determines
    whether an existing file will be read or a new file will be created. Valid
    values are ``"r"`` for reading a file and ``"w"`` for writing a file.
     
    Returns a :class:`Media_Read` or :class:`Media_Write` object representing
    the open file.
    """
    global _av_is_registered
    if not _av_is_registered:
        avformat.av_register_all()
        _av_is_registered = True

    if mode=="r":
        return Media_Read_ffmpeg(name, **keyargs)
    elif mode=="w":
        return Media_Write_ffmpeg(name, **keyargs)
    else:
        raise ValueError("Unknown mode '%s' (must be 'r' or 'w')"%mode)
        



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: quat.py,v 1.1 2005/08/15 15:39:48 mbaas Exp $

import math
from .vec3 import vec3 as _vec3
from .mat3 import mat3 as _mat3
from .mat4 import mat4 as _mat4

# Comparison threshold
_epsilon = 1E-12

# quat
class quat:
    """Quaternion class.

    Quaternions are an extension to complex numbers and can be used
    to store rotations. They are composed of four floats which can be
    seen as an angle and an axis of rotation.
    """

    def __init__(self, *args):
        """Constructor.

        0 arguments: zeroes
        1 float argument:  w component, x,y,z = (0,0,0)
        1 quat argument: Make a copy
        1 mat3 argument: Initialize by rotation matrix
        1 mat4 argument: Initialize by rotation matrix
        2 arguments: angle & axis (doesn't have to be of unit length)
        4 arguments: components w,x,y,z
        """

        # 0 arguments
        if len(args)==0:
            self.w, self.x, self.y, self.z = (0.0, 0.0, 0.0, 0.0)

        # 1 arguments
        elif len(args)==1:
            T = type(args[0])
            # Scalar
            if T in [float, int]:
                self.w = float(args[0])
                self.x, self.y, self.z = (0.0, 0.0, 0.0)
            # quat
            elif isinstance(args[0], quat):
                q=args[0]
                self.w = q.w
                self.x = q.x
                self.y = q.y
                self.z = q.z
            # mat3 or mat4
            elif isinstance(args[0], _mat3) or isinstance(args[0], _mat4):
                self.fromMat(args[0])
            # List or Tuple
            elif T in [list, tuple]:
                dummy = quat(*args[0])
                self.w = dummy.w
                self.x = dummy.x
                self.y = dummy.y
                self.z = dummy.z                
            # String
            elif T is str:
                s=args[0].replace(","," ").replace("  "," ").strip().split(" ")
                if s==[""]:
                    s=[]
                f=list([float(x) for x in s])
                dummy = quat(f)
                self.w = dummy.w
                self.x = dummy.x
                self.y = dummy.y
                self.z = dummy.z
            else:
                raise TypeError("quat() arg can't be converted to quat")

        # 2 arguments (angle & axis)
        elif len(args)==2:
            angle, axis = args
            self.fromAngleAxis(angle,axis)
            
        # 4 arguments
        elif len(args)==4:
            w,x,y,z = args
            self.w = float(w)
            self.x = float(x)
            self.y = float(y)
            self.z = float(z)

        else:
            raise TypeError("quat() arg can't be converted to quat")
        

    def __repr__(self):
        return 'quat(%r, %r, %r, %r)'%(self.w, self.x, self.y, self.z)

    def __str__(self):
        return '(%1.4f, %1.4f, %1.4f, %1.4f)'%(self.w, self.x, self.y, self.z)

    def __eq__(self, other):
        """== operator

        >>> a=quat(1,2,3,4)
        >>> b=quat(6,7,8,9)
        >>> c=quat(6,7,8,9)
        >>> print a==b
        0
        >>> print b==c
        1
        >>> print a==None
        0
        """
        global _epsilon
        if isinstance(other, quat):
            return (abs(self.x-other.x)<=_epsilon and
                    abs(self.y-other.y)<=_epsilon and
                    abs(self.z-other.z)<=_epsilon and
                    abs(self.w-other.w)<=_epsilon)
        else:
            return False

    def __ne__(self, other):
        """!= operator

        >>> a=quat(1,2,3,4)
        >>> b=quat(6,7,8,9)
        >>> c=quat(6,7,8,9)
        >>> print a!=b
        1
        >>> print b!=c
        0
        >>> print a!=None
        1
        """
        return not (self==other)

    def __add__(self, other):
        """Addition.

        >>> q=quat(0.9689, 0.2160, 0.1080, 0.0540)
        >>> print q+q
        (1.9378, 0.4320, 0.2160, 0.1080)
        """
        if isinstance(other, quat):
            return quat(self.w+other.w, self.x+other.x,
                        self.y+other.y, self.z+other.z)
        else:
            raise TypeError("unsupported operand type for +")

    def __sub__(self, other):
        """Subtraction.

        >>> q=quat(0.9689, 0.2160, 0.1080, 0.0540)
        >>> print q-q
        (0.0000, 0.0000, 0.0000, 0.0000)
        """
        if isinstance(other, quat):
            return quat(self.w-other.w, self.x-other.x,
                        self.y-other.y, self.z-other.z)
        else:
            raise TypeError("unsupported operand type for +")

    def __mul__(self, other):
        """Multiplication.

        >>> q=quat(0.9689, 0.2160, 0.1080, 0.0540)
        >>> print q*2.0
        (1.9378, 0.4320, 0.2160, 0.1080)
        >>> print 2.0*q
        (1.9378, 0.4320, 0.2160, 0.1080)
        >>> print q*q
        (0.8775, 0.4186, 0.2093, 0.1046)
        """
        T = type(other)
        # quat*scalar
        if T in [float, int]:
            return quat(self.w*other, self.x*other, self.y*other, self.z*other)
        # quat*quat
        if isinstance(other, quat):
            w1,x1,y1,z1 = self.w,self.x,self.y,self.z
            w2,x2,y2,z2 = other.w,other.x,other.y,other.z
            return quat(w1*w2-x1*x2-y1*y2-z1*z2,
                        w1*x2+x1*w2+y1*z2-z1*y2,
                        w1*y2+y1*w2-x1*z2+z1*x2,
                        w1*z2+z1*w2+x1*y2-y1*x2)
        # unsupported
        else:
            # Try to delegate the operation to the other operand
            if getattr(other,"__rmul__",None)!=None:
                return other.__rmul__(self)
            else:
                raise TypeError("unsupported operand type for *")

    __rmul__ = __mul__

    def __truediv__(self, other):
        """Division.

        >>> q=quat(0.9689, 0.2160, 0.1080, 0.0540)
        >>> print q/2.0
        (0.4844, 0.1080, 0.0540, 0.0270)
        """
        T = type(other)
        # quat/scalar
        if T in [float, int]:
            return quat(self.w/other, self.x/other, self.y/other, self.z/other)
        # unsupported
        else:
            raise TypeError("unsupported operand type for /")
        
    def __pow__(self, other):
        """Return self**q."""
#        if modulo!=None:
#            raise TypeError, "unsupported operation"
        q = quat(other)
        return (q*self.log()).exp()

    def __neg__(self):
        """Negation.

        >>> q=quat(0.9689, 0.2160, 0.1080, 0.0540)
        >>> print -q
        (-0.9689, -0.2160, -0.1080, -0.0540)
        """
        return quat(-self.w, -self.x, -self.y, -self.z)

    def __pos__(self):
        """
        >>> q=quat(0.9689, 0.2160, 0.1080, 0.0540)
        >>> print +q
        (0.9689, 0.2160, 0.1080, 0.0540)
        """
        return quat(+self.w, +self.x, +self.y, +self.z)

    def __abs__(self):
        """Return magnitude.
        
        >>> q=quat(0.9689, 0.2160, 0.1080, 0.0540)
        >>> print round(abs(q),5)
        1.0
        """
        return math.sqrt(self.w*self.w + self.x*self.x +
                         self.y*self.y + self.z*self.z)

    def conjugate(self):
        """Return conjugate.
        
        >>> q=quat(0.9689, 0.2160, 0.1080, 0.0540)
        >>> print q.conjugate()
        (0.9689, -0.2160, -0.1080, -0.0540)
        """
        return quat(self.w, -self.x, -self.y, -self.z)

    def normalize(self):
        """Return normalized quaternion.

        >>> q=quat(0.9, 0.5, 0.2, 0.3)
        >>> q=q.normalize()
        >>> print q
        (0.8250, 0.4583, 0.1833, 0.2750)
        >>> print abs(q)
        1.0
        """
        nlen = 1.0/abs(self)
        return quat(self.w*nlen, self.x*nlen, self.y*nlen, self.z*nlen)

    def inverse(self):
        """Return inverse.

        >>> q=quat(0.9, 0.5, 0.2, 0.3)
        >>> print q.inverse()
        (0.7563, -0.4202, -0.1681, -0.2521)
        """
        len_2 = self.w*self.w + self.x*self.x + self.y*self.y + self.z*self.z
        return self.conjugate()/len_2

    def toAngleAxis(self):
        """Return angle (in radians) and rotation axis.

        >>> q=quat(0.9, 0.5, 0.2, 0.3)
        >>> angle, axis = q.toAngleAxis()
        >>> print round(angle,4)
        1.2011
        >>> print axis
        (0.8111, 0.3244, 0.4867)
        """

        nself = self.normalize()
        
        # Clamp nself.w (since the quat has to be normalized it should
        # be between -1 and 1 anyway, but it might be slightly off due
        # to numerical inaccuracies)
        w = max(min(nself.w,1.0),-1.0)
        
        w = math.acos(w)
        s = math.sin(w)
        if s<1E-12:
            return (0.0, _vec3(0.0,0.0,0.0))
        return (2.0*w, _vec3(nself.x/s, nself.y/s, nself.z/s))

    def fromAngleAxis(self, angle, axis):
        """Initialize self from an angle (in radians) and an axis and returns self."""
        if axis==_vec3(0):
            self.w = 1.0
            self.x = 0.0
            self.y = 0.0
            self.z = 0.0
        else:
            angle/=2.0
            self.w = math.cos(angle)
            x, y, z = axis
            s = math.sin(angle)/math.sqrt(x*x+y*y+z*z)
            self.x = x*s
            self.y = y*s
            self.z = z*s
            dummy = self.normalize()
            self.w = dummy.w
            self.x = dummy.x
            self.y = dummy.y
            self.z = dummy.z

        return self

    def toMat3(self):
        """Return rotation matrix as mat3."""
        x,y,z,w = self.x, self.y, self.z, self.w
        xx = 2.0*x*x
        yy = 2.0*y*y
        zz = 2.0*z*z
        xy = 2.0*x*y
        zw = 2.0*z*w
        xz = 2.0*x*z
        yw = 2.0*y*w
        yz = 2.0*y*z
        xw = 2.0*x*w
        return _mat3(1.0-yy-zz, xy-zw, xz+yw,
                     xy+zw, 1.0-xx-zz, yz-xw,
                     xz-yw, yz+xw, 1.0-xx-yy)

    def toMat4(self):
        """Return rotation matrix as mat4."""
        x,y,z,w = self.x, self.y, self.z, self.w
        xx = 2.0*x*x
        yy = 2.0*y*y
        zz = 2.0*z*z
        xy = 2.0*x*y
        zw = 2.0*z*w
        xz = 2.0*x*z
        yw = 2.0*y*w
        yz = 2.0*y*z
        xw = 2.0*x*w
        return _mat4(1.0-yy-zz, xy-zw, xz+yw, 0.0,
                     xy+zw, 1.0-xx-zz, yz-xw, 0.0,
                     xz-yw, yz+xw, 1.0-xx-yy, 0.0,
                     0.0, 0.0, 0.0, 1.0)

    def fromMat(self, m):
        """Initialize self from either a mat3 or mat4 and returns self."""
        global _epsilon
        
        d1,d2,d3 = m[0,0],m[1,1],m[2,2]
        t = d1+d2+d3+1.0
        if t>_epsilon:
            s = 0.5/math.sqrt(t)
            self.w = 0.25/s
            self.x = (m[2,1]-m[1,2])*s
            self.y = (m[0,2]-m[2,0])*s
            self.z = (m[1,0]-m[0,1])*s
        else:
            ad1 = d1
            ad2 = d2
            ad3 = d3
            if ad1>=ad2 and ad1>=ad3:
                s = math.sqrt(1.0+d1-d2-d3)*2.0
                self.x = 0.25*s
                self.y = (m[0,1]+m[1,0])/s
                self.z = (m[0,2]+m[2,0])/s
                self.w = (m[1,2]+m[2,1])/s
            elif ad2>=ad1 and ad2>=ad3:
                s = math.sqrt(1.0+d2-d1-d3)*2.0
                self.x = (m[0,1]+m[1,0])/s
                self.y = 0.25*s
                self.z = (m[1,2]+m[2,1])/s
                self.w = (m[0,2]+m[2,0])/s
            else:
                s = math.sqrt(1.0+d3-d1-d2)*2.0
                self.x = (m[0,2]+m[2,0])/s
                self.y = (m[1,2]+m[2,1])/s
                self.z = 0.25*s
                self.w = (m[0,1]+m[1,0])/s

        return self

    def dot(self, b):
        """Return the dot product of self and b."""
        return self.w*b.w + self.x*b.x + self.y*b.y + self.z*b.z

    def log(self):
        """Return the natural logarithm of self."""
        global _epsilon
        
        b = math.sqrt(self.x*self.x + self.y*self.y + self.z*self.z)
        res = quat()
        if abs(b)<=_epsilon:
            res.x = 0.0
            res.y = 0.0
            res.z = 0.0
            if self.w<=_epsilon:
                raise ValueError("math domain error")
            res.w = math.log(self.w)
        else:
            t = math.atan2(b, self.w)
            f = t/b
            res.x = f*self.x
            res.y = f*self.y
            res.z = f*self.z
            ct = math.cos(t)
            if abs(ct)<=_epsilon:
                raise ValueError("math domain error")
            r = self.w/ct
            if r<=_epsilon:
                raise ValueError("math domain error")
            res.w = math.log(r)

        return res

    def exp(self):
        """Return the exponential of self."""
        global _epsilon
        
        b = math.sqrt(self.x*self.x + self.y*self.y + self.z*self.z)
        res = quat()
        if abs(b)<=_epsilon:
            res.x = 0.0
            res.y = 0.0
            res.z = 0.0
            res.w = math.exp(self.w)
        else:
            f = math.sin(b)/b
            res.x = f*self.x
            res.y = f*self.y
            res.z = f*self.z
            res.w = math.exp(self.w)*math.cos(b)

        return res

    def rotateVec(self, v):
        """Return the rotated vector v.

        The quaternion must be a unit quaternion.
        This operation is equivalent to turning v into a quat, computing
        self*v*self.conjugate() and turning the result back into a vec3.
        """

        v = _vec3(v)
        ww = self.w*self.w
        xx = self.x*self.x
        yy = self.y*self.y
        zz = self.z*self.z
        wx = self.w*self.x
        wy = self.w*self.y
        wz = self.w*self.z
        xy = self.x*self.y
        xz = self.x*self.z
        yz = self.y*self.z

        return _vec3(ww*v.x + xx*v.x - yy*v.x - zz*v.x + 2*((xy-wz)*v.y + (xz+wy)*v.z),
                     ww*v.y - xx*v.y + yy*v.y - zz*v.y + 2*((xy+wz)*v.x + (yz-wx)*v.z),
                     ww*v.z - xx*v.z - yy*v.z + zz*v.z + 2*((xz-wy)*v.x + (yz+wx)*v.y))
    

def slerp(t, q0, q1, shortest=True):
    """Spherical linear interpolation between two quaternions.

    The return value is an interpolation between q0 and q1. For t=0.0
    the return value equals q0, for t=1.0 it equals q1.
    q0 and q1 must be unit quaternions.
    If shortest is True the interpolation is always done along the
    shortest path.
    """
    global _epsilon

    ca = q0.dot(q1)
    if shortest and ca<0:
        ca = -ca
        neg_q1 = True
    else:
        neg_q1 = False
    
    if ca>=1.0:
        o = 0.0
    elif ca<=-1.0:
        o = math.pi
    else:
        o = math.acos(ca)
    so = math.sin(o)

    if (abs(so)<=_epsilon):
        return quat(q0)

    a = math.sin(o*(1.0-t)) / so
    b = math.sin(o*t) / so
    if neg_q1:
        return q0*a - q1*b
    else:
        return q0*a + q1*b

def squad(t, a, b, c, d):
    """Spherical cubic interpolation."""
    return slerp(2*t*(1.0-t), slerp(t,a,d), slerp(t,b,c))


######################################################################

def _test():
    import doctest, quat
    failed, total = doctest.testmod(quat)
    print(("%d/%d failed" % (failed, total)))

if __name__=="__main__":

    _test()

#    q = quat(1.5,_vec3(1,0,0))
#    print q
#    m=q.toMat4().getMat3()
#    print m
#    w=quat(m)
#    print w





# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: __init__.py,v 1.19 2006/02/14 19:29:39 mbaas Exp $

"""Python Computer Graphics Kit
============================

The Python Computer Graphics Kit is a generic 3D package that can be
useful in any domain where you have to deal with 3D data of any kind,
be it for visualization, creating photorealistic images, Virtual
Reality or even games.

At its lowest level, the package provides the basic functionality that
is useful for writing your own tools that process 3D data. For
example, the cgtypes module defines the fundamental types for computer
graphics such as vectors and matrices, the ri module contains the
complete RenderMan API to create RIB files, and so on.

Using these relatively low level modules the second level provides the
functionality to store a 3D scene in memory. This is the major part of
the package and this actually turns the general purpose language
Python into a specialized scripting language like MEL or MaxScript,
for example.

Eventually, the package provides small tools that let you actually see
your 3D scene. The two standard tools are for interactive rendering
using OpenGL and for offline rendering using a RenderMan renderer.

See the manual and tutorials at: http://cgkit.sourceforge.net/
"""

import os.path
from . import cgkitinfo

# Is this cgkit light? Then modify the subpackage searchpath so that the
# special 'light' modules are found instead of the normal ones.
if cgkitinfo.cgkit_light:
    # Package's main folder
    dirname = __path__[0]
    # Put the 'light' subpackage in front so that its contents shadows
    # the normal extension modules...
    __path__.insert(0, os.path.join(dirname, "light"))



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2010
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import ctypes
from . import findlib
from . import decls
from .decls import AVRational

class AVError(Exception):
    """Base AV error class.
    """
    def __init__(self, msgOrErrnum):
        """Constructor.
        
        msgOrErrnum is either a string containing an error message or an integer
        containing the ffmpeg error number. If a number is passed, the number
        is converted into an error message.
        The error number is stored in the attribute errnum. If a string has
        been passed, that attribute will be None.
        """
        self.errnum = None
        # If the input argument is an int, then try to convert it into a message
        if type(msgOrErrnum) is int:
            self.errnum = msgOrErrnum
            msg = av_strerror(self.errnum)
            if msg is None:
                msg = "Error %s"%(msgOrErrnum)
        else:
            msg = msgOrErrnum
        
        Exception.__init__(self, msg)


def avutil_version():
    """Return the libavutil library version.
    
    Returns a tuple (major,minor,micro) containing the three parts of the
    version number. 
    """
    v = _lib().avutil_version()
    major = v>>16
    minor = (v>>8)&0xff
    micro = v&0xff
    return (major,minor,micro)

def av_free(obj):
    """Free memory which has been allocated with av_malloc(z)() or av_realloc().
    
    *obj* may be a AVFrame object that was allocated using avcodec_alloc_frame()
    (note: *obj* must be the object itself, not a pointer to it).
    """
    if obj is None:
        return
    _lib().av_free(ctypes.byref(obj))

def av_d2q(d, max):
    """Convert a floating point number into a rational.
    
    *d* is the floating point number and *max* the maximum allowed numerator
    and denominator
    Returns a :class:`AVRational` object representing *d*.
    """
    func = _lib().av_d2q
    func.args = [ctypes.c_double, ctypes.c_int]
    func.restype = AVRational
    return func(ctypes.c_double(d), int(max))

def av_strerror(errnum):
    """Return a description of an AVERROR code.
    
    Returns ``None`` if no description could be found. 
    """
    buf = ctypes.create_string_buffer(200)
    func = _lib().av_strerror
    res = func(int(errnum), buf, ctypes.sizeof(buf))
    if res==0:
        return buf.value
    else:
        return None

def av_get_sample_fmt_name(sampleFmt):
    """Return the name of the sample format or ``None`` if an unknown value is passed.
    
    *sampleFmt* is an ``AV_SAMPLE_FMT_*`` value.
    """
    func = _lib().av_get_sample_fmt_name
    func.restype = ctypes.c_char_p
    name = func(sampleFmt)
    return name

def av_get_sample_fmt(name):
    """Return the sample format corresponding to the given name.
    
    Returns ``AV_SAMPLE_FMT_NONE`` if the name is not recognized.
    """
    return _lib().av_get_sample_fmt(name)

def av_get_bytes_per_sample(sampleFmt):
    """Return the number of bytes per sample.
    
    *sampleFmt* is an ``AV_SAMPLE_FMT_*`` value. Returns 0 if sampleFmt refers to
    an unknown sample format.
    """
    return _lib().av_get_bytes_per_sample(sampleFmt)

def av_get_pix_fmt_name(pixFmt):
    """Return the name of the pixel format or ``None`` if an unknown value is passed.
    
    *pixFmt* is a ``PIX_FMT_*`` value.
    """
    func = _lib().av_get_pix_fmt_name
    func.restype = ctypes.c_char_p
    name = func(pixFmt)
    return name
    
def av_get_pix_fmt(name):
    """Return the pixel format corresponding to the given name.
    
    Returns ``PIX_FMT_NONE`` if the name is not recognized.
    """
    return _lib().av_get_pix_fmt(name)


# By default, try to load the avutil library that has the same major version
# than the one that was used for creating the cppdefs and decls module.
_libname = "avutil.%s"%(decls.LIBAVUTIL_VERSION_MAJOR)
_libavutil = None

def _lib():
    """Return the avutil shared library.
    """
    global _libavutil
    global _libname
    
    if _libavutil is None:
        _libavutil = findlib.findFfmpegLib(_libname)
    return _libavutil

